--- Confetti

class("Confetti")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local PARTICLE_COLORS = {
  {99, 180, 211},
  {91, 184, 129},
  {229, 111, 45},
  {246, 197, 68}
}

local Z_INDEX = {
  CIRCLES = 10,
  SQUARES = 20
}

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function makeCircleParticle()
  local minRadius, maxRadius = 1, 2
  local radius = math.random(minRadius, maxRadius)
  local color = ColorUtils.colorRGBToStd(unpack(TableUtils.getRandomElement(PARTICLE_COLORS)))
  local circle = FilledCircle.new(radius, color)
  local group = Group.new(radius*2, radius*2)
  group:insertAtLocationFractional(circle, math.random(), math.random())
  group:linkOpacityAttrToChildren()
  group:setOpacity(0)
  return group
end

local function getCircleParticleAnimation(self, circle, duration, blastDuration, tailDuration, fadeDuration, isLeft)
  local heightAmount = self.emitY + (math.random() * (1 - self.emitY))
  local widthAmount = math.random() * 0.8
  if isLeft then
    widthAmount = -widthAmount
  end
  local fallDistance = self:getWidthPart(0.4 * math.random())
  local moveOutEase = AHEaseExponentialOut.new()
  local fallEase = AHEaseLinear.new()
  local moveOutX = AHLocationXBy.new(circle, blastDuration, moveOutEase, self:getWidthPart(widthAmount))
  local moveOutY = AHLocationYTo.new(circle, blastDuration, moveOutEase, self:getHeightPart(heightAmount))
  local moveDownY = AHLocationYTo.new(circle, duration, fallEase, -fallDistance)
  local moveSequence = AHSpawn.new(moveOutY, moveDownY)
  local rotateAnim = AHRotationBy.new(circle, tailDuration, fallEase, math.random() * 720)
  local fadeAnim = AHOpacityTo.new(circle, fadeDuration, fallEase, 0)
  local particleAnim = AHSpawn.new(moveOutX, moveSequence)
  particleAnim:addAction(AHDelayed.new(blastDuration, rotateAnim))
  particleAnim:addAction(AHDelayed.new((duration - fadeDuration), fadeAnim))
  particleAnim:addAction(function() circle:setOpacity(1) end)
  return particleAnim
end

local function makeSquareParticle(makeLargeParticle)
  local minSize, maxSize = 12, 30
  if makeLargeParticle then
    minSize, maxSize = 20, 60
  end
  local size = math.random(minSize, maxSize)
  local color = ColorUtils.colorRGBToStd(unpack(TableUtils.getRandomElement(PARTICLE_COLORS)))
  local square = FilledRectangle.new(size, size, color)
  local group = Group.new(size, size)
  group:insertAtLocationFractional(square, math.random(), math.random())
  group:linkOpacityAttrToChildren()
  group:setOpacity(0)
  return group
end

local function getSquareParticleAnimation(self, square, duration, blastDuration, tailDuration, fadeDuration, isLeft)
  local heightAmount = self.emitY + (math.random() * (1 - self.emitY))
  local widthAmount = math.random() * 0.8
  if isLeft then
    widthAmount = -widthAmount
  end
  if self.emitFromCenter then
    widthAmount = widthAmount * 0.8
  end
  local fallDistance = self:getWidthPart(0.6 * math.random())
  local moveOutEase = AHEaseExponentialOut.new()
  local moveOutX = AHLocationXBy.new(square, blastDuration, moveOutEase, self:getWidthPart(widthAmount))
  local moveOutY = AHLocationYTo.new(square, blastDuration, moveOutEase, self:getHeightPart(heightAmount))
  local moveDownY = AHLocationYTo.new(square, duration, AHEaseLinear.new(), -fallDistance)
  local moveSequence = AHSpawn.new(moveOutY, moveDownY)
  local offsetX, offsetY = math.random(), math.random()
  local rotateUpdateFunction = function(normalizedTime)
    local timeFactor = normalizedTime * 1.5
    square:setScl(math.sin(timeFactor + offsetX), math.cos(timeFactor + offsetY))
  end
  rotateUpdateFunction(0)
  square:setRot(math.random() * 90)
  local rotateAnim = AHCustomAction.new(tailDuration, AHEaseLinear.new(), rotateUpdateFunction)
  local fadeAnim = AHOpacityTo.new(square, fadeDuration, AHEaseLinear.new(), 0)
  local particleAnim = AHSpawn.new(moveOutX, moveSequence)
  particleAnim:addAction(AHDelayed.new(blastDuration, rotateAnim))
  particleAnim:addAction(AHDelayed.new((duration - fadeDuration), fadeAnim))
  particleAnim:addAction(function() square:setOpacity(1) end)
  return particleAnim
end

local function getParticleAnimation(self, numSquares, numCircles, duration, isLeft)
  local blastDuration = duration * 0.35
  local tailDuration = duration - blastDuration
  local fadeDuration = tailDuration * 0.2
  local emitX = 0
  if isLeft then
    emitX = 1
  end
  if self.emitFromCenter then
    emitX = 0.5
  end
  local fullAnim = AHSpawn.new()
  for i=1, numCircles do
    local circle = makeCircleParticle()
    self:insertAtLocationFractional(circle, emitX, self.emitY, Z_INDEX.CIRCLES)
    local particleAnim = getCircleParticleAnimation(
      self, circle, duration, blastDuration, tailDuration, fadeDuration, isLeft)
    fullAnim:addAction(particleAnim)
  end
  for i=1, numSquares do
    local square = makeSquareParticle(self.emitFromCenter)
    self:insertAtLocationFractional(square, emitX, self.emitY, Z_INDEX.SQUARES)
    local particleAnim = getSquareParticleAnimation(
      self, square, duration, blastDuration, tailDuration, fadeDuration, isLeft)
    fullAnim:addAction(particleAnim)
  end
  return fullAnim
end

local function getConfettiAnimation(self)
  local numSquares = 15
  local numCircles = 10
  local duration = 2.5
  if self.emitFromCenter then
    numSquares = 20
    numCircles = 0
    duration = 3.5
  end
  local leftAnim = getParticleAnimation(self, numSquares, numCircles, duration, true)
  local rightAnim = getParticleAnimation(self, numSquares, numCircles, duration, false)
  return AHSpawn.new(leftAnim, rightAnim)
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function Confetti:getAnimation()
  if #self.loadedAnimations == 0 then
    self:loadAnimations(1)
  end
  return table.remove(self.loadedAnimations)
end

function Confetti:loadAnimations(numToLoad)
  for i=1, numToLoad do
    table.insert(self.loadedAnimations, getConfettiAnimation(self))
  end
end

function Confetti:setEmitHeight(height)
  self.emitY = height
end

function Confetti:setEmitFromCenter(fromCenter)
  self.emitFromCenter = fromCenter
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

-- Constructor ------------------------

function Confetti.new(width, height, emitHeight)
  local self = Group.new(width, height)
  Mixin.mix(self, Confetti)

  self.emitY = emitHeight

  self.emitFromCenter = false

  self.loadedAnimations = {}

  return self
end
