--- GameResultScreen

class("GameResultScreen")

require("sphinx.RectangleButton")
require("sphinx.SphinxColors")
require("sphinx.SphinxTimeDisplay")


-- ----------------------------------------------------------------------------
-- Stat Pairs
-- ----------------------------------------------------------------------------

local DIFFICULTY_STRING = "Difficulty"
local HINT_STRING = "Hints"

local SCREEN_NAME = "game_result"

function GameResultScreen.getCustomStat(title, description)
  return {
    title = title,
    description = description
  }
end

function GameResultScreen.getXPEarned()
  return GameResultScreen.getCustomStat(Localizer.get("Sphinx.XP_EARNED"), 25)
end

function GameResultScreen.getHintCountStat(hintCount)
  return GameResultScreen.getCustomStat(HINT_STRING, hintCount)
end

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local ROUNDS_COMPLETED_TEXT = "%d Rounds Completed"
local TAP_TO_CONTINUE_TEXT = "TAP TO CONTINUE"

local Z_INDEX = {
  UI = 0,
  STARS = 10,
  BUTTON_OVERLAY = 20
}

local STATS_MOVE_DISTANCE = ReferenceDevice.getScaledHeight(30)

local CONGRATULATORY_MESSAGE_STRINGS = {
  {
    "Practice makes perfect!",
    "Way to go!",
    "You're getting there!",
    "You did it!",
    "Nice try!",
    "Better luck next time!",
    "You gave it your best!",
    "Slow and steady wins the race!"
  },
  {
    "Great job!",
    "Solid work!",
    "Very nice!",
    "Nicely done!",
    "Well done!",
    "You're picking up speed!",
    "You're getting the hang of this!",
    "Coming along nicely!"
  },
  {
    "Stellar!!",
    "Stupendous!!",
    "Incredible!!",
    "Magnificent!!",
    "Marvelous!!",
    "Sublime!!",
    "Fantastic!",
    "Astounding!!",
    "You rock!",
    "You rule!",
    "You're the best!",
    "Masterfully done!",
    "Superb puzzle-solving skills!",
    "That's how it's done!",
    "A magnificent performance!",
    "Supercalifragilisticexpialidocious!"
  }
}

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function createTitleGroup(self)
  local gameName = Label.new(self.displayName, Font.new(FontFaces.BOLD, 24))
  local rankReplacement = 3 - MathUtils.clamp(0, 2, math.floor((self.endTime / 60)))
  local messageString = TableUtils.getRandomElement(CONGRATULATORY_MESSAGE_STRINGS[rankReplacement])
  local congratulatoryMessage = Label.new(messageString, Font.new(FontFaces.BOLD, 18))

  local spacing = ReferenceDevice.getScaledHeight(7.5)
  local titleGroup = Column.new(spacing)
  titleGroup:insert(gameName)
  titleGroup:insert(congratulatoryMessage)
  titleGroup:alignChildren()

  titleGroup.title = gameName
  congratulatoryMessage:setColor(unpack(SphinxColors.get("primaries", "yellow")))
  gameName:setOpacity(0)
  congratulatoryMessage:setOpacity(0)

  local duration = 0.3
  local ease = nil
  titleGroup.showGameNameAnimation = AHOpacityTo.new(gameName, duration, ease, 1)
  titleGroup.showMessageAnimation = AHOpacityTo.new(congratulatoryMessage, duration, ease, 1)

  titleGroup.hideGameNameAnimation = AHOpacityTo.new(gameName, duration, ease, 0)
  titleGroup.hideMessageAnimation = AHOpacityTo.new(congratulatoryMessage, duration, ease, 0)

  return titleGroup
end

local function createTimeGroup(self)
  local timeGroup = Column.new()
  local timeSolvedText = Localizer.get("Sphinx.COMPLETION_TIME")
  local subtitleLabel = Label.new(timeSolvedText, Font.new(FontFaces.SEMI_BOLD, 24))
  timeGroup:insert(subtitleLabel)
  timeGroup:insertSpacer(-4)
  local timerFont = Font.new(FontFaces.BOLD, 80)
  local timerColor = Colors.WHITE
  local timerLabel = SphinxTimeDisplay.new(timerFont, timerColor)
  timerLabel:setDisplayedTime(self.endTime)
  timeGroup:insert(timerLabel)
  timeGroup:alignChildren()

  local isRecursive = true
  timeGroup:linkOpacityAttrToChildren(isRecursive)
  timeGroup:setOpacity(0)

  local duration = 0.3
  local ease = nil
  timeGroup.showAnimation = AHOpacityTo.new(timeGroup, duration, ease, 1.0)
  timeGroup.hideAnimation = AHOpacityTo.new(timeGroup, duration, ease, 0.0)

  return timeGroup
end

local function makeStatGroup(title, description, width)
  local titleLabel = Label.new(title, Font.new(FontFaces.BOLD, 18))
  titleLabel.maxOpacity = 0.5
  local descriptionLabel = Label.new(description, Font.new(FontFaces.SEMI_BOLD, 20))
  local statGroup = Row.new()
  statGroup:setWidth(width)
  statGroup:insert(titleLabel)
  statGroup:insert(descriptionLabel)
  statGroup:alignChildren()

  titleLabel:setOpacity(0)
  descriptionLabel:setOpacity(0)

  local showAnimation = AHSpawn.new()
  local fadeInDuration = 0.75
  local fadeInEase = AHEaseBezier.newFlat({0.26, 1, 0.48, 1})
  showAnimation:addAction(AHOpacityTo.new(titleLabel, fadeInDuration, fadeInEase, titleLabel.maxOpacity))
  showAnimation:addAction(AHOpacityTo.new(descriptionLabel, fadeInDuration, fadeInEase, 1))
  showAnimation:addAction(AHLocationYBy.new(titleLabel, fadeInDuration, fadeInEase, STATS_MOVE_DISTANCE))
  showAnimation:addAction(AHLocationYBy.new(descriptionLabel, fadeInDuration, fadeInEase, STATS_MOVE_DISTANCE))
  statGroup.showAnimation = showAnimation

  local hideAnimation = AHSpawn.new()
  local fadeOutDuration = 0.283
  local fadeOutEase = AHEaseLinear.new()
  hideAnimation:addAction(AHOpacityTo.new(titleLabel, fadeOutDuration, fadeOutEase, 0))
  hideAnimation:addAction(AHOpacityTo.new(descriptionLabel, fadeOutDuration, fadeOutEase, 0))
  statGroup.hideAnimation = hideAnimation

  return statGroup
end

local function createStatsGroup(self)
  local width = ReferenceDevice.getScaledWidth(275)
  local statGroups = {}
  for _, statPair in ipairs(self.stats) do
    local title = statPair.title
    local description = statPair.description
    table.insert(statGroups, makeStatGroup(title, description, width))
  end

  local getLine = function()
    local line = FilledRectangle.new(width, 2, {1,1,1})
    line.maxOpacity = 0.1
    line:setOpacity(0)
    local fadeInDuration = 0.75
    local fadeInEase = AHEaseBezier.newFlat({0.26, 1, 0.48, 1})
    line.showAnimation = AHSpawn.new(
      AHOpacityTo.new(line, fadeInDuration, fadeInEase, line.maxOpacity),
      AHLocationYBy.new(line, fadeInDuration, fadeInEase, STATS_MOVE_DISTANCE))

    local fadeOutDuration = 0.283
    local fadeOutEase = AHEaseLinear.new()
    line.hideAnimation = AHOpacityTo.new(line, fadeOutDuration, fadeOutEase, 0)
    return line
  end

  local spacing = ReferenceDevice.getScaledHeight(14)

  local stats = Column.new(spacing)
  stats.showAnimation = AHSpawn.new()
  stats.hideAnimation = AHSpawn.new()
  local statDelay = 0.25
  for i, statGroup in ipairs(statGroups) do
    local delay = (i - 1) * statDelay
    if i > 1 then
      local line = getLine()
      stats:insert(line)
      stats.showAnimation:addAction(AHDelayed.new(delay, line.showAnimation))
      stats.hideAnimation:addAction(line.hideAnimation)
    end
    stats:insert(statGroup)
    stats.showAnimation:addAction(AHDelayed.new(delay, statGroup.showAnimation))
    stats.hideAnimation:addAction(statGroup.hideAnimation)
  end

  stats:alignChildren()

  return stats
end

local function createContinueButton(self)
  local button = nil
  button = Button.new(Screen.safeWidth, Screen.safeHeight, nil, function()
    button:setTouchEnabled(false)
    self.delegate:moveToNextScreen()
  end)

  button:setTouchEnabled(false)

  local buttonText = Label.new(TAP_TO_CONTINUE_TEXT, PrerenderedFont.new(FontFaces.MEDIUM, 14), nil, nil, nil, 0.8)
  buttonText.maxOpacity = 0.5
  buttonText:setOpacity(0)
  button:insertAtLocationFractional(buttonText, 0.5, 0.07)

  local showAnimation = AHSequence.new()
  showAnimation:addAction(AHOpacityTo.new(buttonText, 0.75, AHEaseBezier.newFlat({0.26, 1, 0.48, 1}), buttonText.maxOpacity))
  showAnimation:addAction(function() button:setTouchEnabled(true) end)

  button.showAnimation = showAnimation

  local hideAnimation = AHSequence.new()
  hideAnimation:addAction(AHOpacityTo.new(buttonText, 0.283, nil, 0))

  button.hideAnimation = hideAnimation

  return button
end

local function createLayout(self)
  self.titleGroup = createTitleGroup(self)
  self.timeGroup = createTimeGroup(self)
  self.statsGroup = createStatsGroup(self)
  self.continueButton = createContinueButton(self)

  self:insertAtCenter(self.continueButton, Z_INDEX.BUTTON_OVERLAY)

  local elements = Column.new()
  elements:setHeight(self:getHeight() - ReferenceDevice.getScaledHeight(62))

  elements:insertSpacer(0)
  elements:insert(self.titleGroup)
  elements:insertSpacer(0)
  elements:insert(self.timeGroup)
  elements:insertSpacer(0)
  elements:insert(self.statsGroup)
  elements:insertSpacer(0)
  elements:insertSpacer(0)
  elements:insertSpacer(0)
  elements:alignChildren()

  self.statsGroup:offsetY(-STATS_MOVE_DISTANCE)

  self:insertAtCenter(elements, Z_INDEX.STARS)
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function GameResultScreen:getTitleYPos()
  local _, locY = self:transformLoc(self.titleGroup.title, self.titleGroup.title:getLoc())
  return locY
end

function GameResultScreen:show()
  --Analytics.sendScreenViewEvent(SCREEN_NAME)
  createLayout(self)

  local delayOffset = 0.3
  local animation = AHSpawn.new()
  animation:addAction(self.titleGroup.showGameNameAnimation)
  animation:addAction(AHDelayed.new(delayOffset, self.titleGroup.showMessageAnimation))
  animation:addAction(AHDelayed.new(delayOffset * 2, self.timeGroup.showAnimation))
  animation:addAction(AHDelayed.new(delayOffset * 3, self.statsGroup.showAnimation))
  animation:addAction(AHDelayed.new(delayOffset * 4, self.continueButton.showAnimation))
  animation:start()
end

function GameResultScreen:hide()
  local animation = AHSpawn.new()
  animation:addAction(self.titleGroup.hideGameNameAnimation)
  animation:addAction(self.titleGroup.hideMessageAnimation)
  animation:addAction(self.timeGroup.hideAnimation)
  animation:addAction(self.statsGroup.hideAnimation)
  animation:addAction(self.continueButton.hideAnimation)
  wait(animation:start())
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

-- Constructor ------------------------

function GameResultScreen.new(delegate, displayName, endTime, gameStats, roundsCompletedCount)
  local self = Group.newFullScreenSafe()
  Mixin.mix(self, GameResultScreen)

  self.delegate = delegate
  self.displayName = displayName
  self.endTime = endTime
  self.stats = gameStats
  self.roundsCompletedCount = roundsCompletedCount

  print(self.endTime)

  return self
end
