--- GenericButton

class("GenericButton")

require("sphinx.SphinxColors")
require("sphinx.DropShadowUtils")


-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local INACTIVE_LABEL_OPACITY = 0.2
local DEFAULT_INACTIVE_COLOR_LIGHTEN = -0.07

local Z_INDEX = {
  SHADOW = 10,
  FACEPLATE = 20,
  LABEL = 30,
  COUNT = 40
}

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function getDownstateAnimation(self)
  local duration = 0.05
  local ease = nil
  return AHLocationTo.new(self.dropShadow, duration, ease, self.faceplate:getXCenter(), self.faceplate:getYCenter())
end

local function getUpstateAnimation(self)
  local duration = 0.05
  local ease = nil
  local x = self.faceplate:getXCenter() + DropShadowUtils.getOffsetX()
  local y = self.faceplate:getYCenter() + DropShadowUtils.getOffsetY()
  return AHLocationTo.new(self.dropShadow, duration, ease, x, y)
end

local function startAnimation(self, animation)
  if self.currentAnimation then
    self.currentAnimation:stop()
  end
  self.currentAnimation = animation
  self.currentAnimation:start()
end

local function getOnTouchDownAction(self)
  return function()
    startAnimation(self, getDownstateAnimation(self))
  end
end

local function getOnTouchUpAction(self, action)
  return function()
    startAnimation(self, getUpstateAnimation(self))
    action(self)
  end
end

local function getOnTouchCancelAction(self)
  return function()
    startAnimation(self, getUpstateAnimation(self))
  end
end

local function setLabelOpacity(self, opacity)
  if self.label then
    self.label:setOpacity(opacity)
  end
end

-- Update the state based on the state flags isHidden and isActive.
-- If isHidden, always hidden. Otherwise, show if active or not based on isActive.
local function updateState(self)
  if self.isVisible then
    self.faceplate:setOpacity(1)
    self.dropShadow:setOpacity(self.dropShadow.maxOpacity)
    if self.isActive then
      self.faceplate:setColor(unpack(self.color))
      setLabelOpacity(self, 1)
      self.super.setTouchEnabled(self, self.isTouchEnabled)
    else
      self.faceplate:setColor(unpack(self.inactiveColor))
      setLabelOpacity(self, INACTIVE_LABEL_OPACITY)
      self.super.setTouchEnabled(self, false)
    end
  else
    self.faceplate:setOpacity(0)
    self.dropShadow:setOpacity(0)
    setLabelOpacity(self, 0)
    self.super.setTouchEnabled(self, false)
  end
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function GenericButton:setButtonColor(color)
  self.color = color
  self.inactiveColor = ColorUtils.lightenColor(color, DEFAULT_INACTIVE_COLOR_LIGHTEN)

  updateState(self)
end

function GenericButton:setCountColor(countColor)
  -- We check for nil to allow calling this method regardless of ads being enabled/disabled.
  if self.countBackground ~= nil then
    self.countBackground:setColor(unpack(countColor))
  end

  updateState(self)
end

function GenericButton:setTextLabel(text)
  if self.label then
    self.label:setText(text)
  end
end

function GenericButton:initializeCount(initialCount, countColor)
  local countLabel = Label.new("", Font.new(FontFaces.BOLD, self.radius * 0.42))
  countLabel:setColor(unpack(SphinxColors.get("primaries", "text")))
  countLabel:inheritOpacity(self.label)

  local countPlayIcon = Sprite.newWithName("play_icon.png")
  countPlayIcon:setEffectiveDims(self.radius * 0.7, nil)
  countPlayIcon:setColor(unpack(unpack(SphinxColors.get("primaries", "text"))))
  countPlayIcon:inheritOpacity(self.label)
  countPlayIcon:setVisible(false)

  local countBackground = FilledCircle.new(self.radius * 0.35, countColor)
  countBackground:inheritOpacity(self.faceplate)
  countBackground:insertAtCenter(countLabel)
  countBackground:insertAtCenter(countPlayIcon)
  countBackground:setAnchorPoint(0.95, 0.35)
  self:insertAtBottomRight(countBackground, Z_INDEX.COUNT)

  self.countLabel = countLabel
  self.countPlayIcon = countPlayIcon
  self.countBackground = countBackground

  self:setCount(initialCount)

  return self
end

function GenericButton:setCount(count)
  assert(self.countLabel ~= nil, "This button does not have a count label")
  assert(count >= 0, "Expected positive count, received: " .. tostring(count))
  self.countLabel:setText(count)
  self.countLabel:setVisible(count > 0)
  self.countPlayIcon:setVisible(count == 0)
end

function GenericButton:setTextColor(color)
  if self.label then
    self.label:setColor(unpack(color))
  end
end

function GenericButton:setVisible(isVisible)
  self.isVisible = isVisible
  updateState(self)
end

function GenericButton:setActive(isActive)
  self.isActive = isActive
  updateState(self)
end

function GenericButton:setTouchEnabled(isEnabled)
  self.isTouchEnabled = isEnabled
  updateState(self)
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

-- Constructor ------------------------

function GenericButton.new(diameter, color, action)
  color = color or {1, 1, 1}

  local self = Group.new(diameter, diameter)
  local touchAcceptedArea = diameter * 1.3
  Button.addVirtualButton(self,
    touchAcceptedArea, touchAcceptedArea,
    getOnTouchDownAction(self),
    getOnTouchUpAction(self, action),
    getOnTouchCancelAction(self))
  Mixin.mix(self, GenericButton)

  self.radius = diameter / 2

  local faceplate = FilledCircle.new(self.radius, color)
  local dropShadow = FilledCircle.new(self.radius, SphinxColors.DROP_SHADOW)

  self:insertAtCenter(dropShadow, Z_INDEX.SHADOW)
  self:insertAtCenter(faceplate, Z_INDEX.FACEPLATE)

  DropShadowUtils.offsetDropShadow(dropShadow, ReferenceDevice.getScaledWidth(54.0))
  dropShadow.maxOpacity = SphinxColors.DROP_SHADOW[4]

  self.faceplate = faceplate
  self.dropShadow = dropShadow

  self.color = color
  self.inactiveColor = ColorUtils.lightenColor(color, DEFAULT_INACTIVE_COLOR_LIGHTEN)

  self.isVisible = true
  self.isActive = true
  self.isTouchEnabled = true

  return self
end

function GenericButton.newWithTextLabel(label, diameter, color, action)
  local self = GenericButton.new(diameter, color, action)

  local textLabel = Label.new(label, Font.new(FontFaces.MEDIUM, self.radius))
  self:insertAtCenter(textLabel, Z_INDEX.LABEL)

  self.label = textLabel

  return self
end

function GenericButton.newWithSpriteLabel(spriteName, diameter, color, action)
  local self = GenericButton.new(diameter, color, action)

  local sprite = Sprite.newWithName(spriteName)
  sprite:setEffectiveDims(self.radius*2)
  self:insertAtCenter(sprite, Z_INDEX.LABEL)

  self.label = sprite

  return self
end

-- ----------------------------------------------------------------------------
-- Simple Buttons
-- ----------------------------------------------------------------------------

function GenericButton.newPauseButton(diameter, color)
  local label = "pause_icon.png"
  local action = function()
    SphinxSound.playClick()
    GameSession.notifyGamePaused()
  end
  local self = GenericButton.newWithSpriteLabel(label, diameter, color, action)
  return self
end
