--- InputDisplay

class("InputDisplay")

require("Event")


-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local PERCENT_PADDING = 0.225
local HEIGHT_TO_CORNER_RADIUS_MULTIPLIER = 0.5
local LETTER_KERNING_FONT_SIZE_MULTIPLER = 0.4

local TIME_WORD_PERSISTS_AFTER_INPUT = 0.2

local DISPLAY_COLOR = {0.0, 0.0, 0.0, 0.5}

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function getBackingWidthForLabel(label)
  return label:getWidth() + label:getHeightPart(1 + PERCENT_PADDING)
end

local function setText(self, text)
  if self.isEffectHappening then return end
  self.label:setText(text)
  local duration = 0.1
  local ease = AHEaseBezier.newFlat({0.26,1,0.48,1})
  if (text == "") then
    self.actionInterrupt:insert(AHSpawn.new(AHOpacityTo.new(self.backing, duration, ease, 0)))
  else
    local width = getBackingWidthForLabel(self.label)
    local widthAnim = AHDimXTo.new(self.backing, duration, ease, width)
    local opacityAnim = AHOpacityTo.new(self.backing, duration, ease, self.backingOpacity)
    self.label:setOpacity(1)
    self.actionInterrupt:insert(AHSpawn.new(widthAnim, opacityAnim))
  end
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function InputDisplay:swipeLetterAdded(letter)
  SphinxSound.playInteractionProgress()
  local currentText = self.text or ""
  self.text = currentText .. letter
  setText(self, self.text)
end

function InputDisplay:swipeLetterRemoved(numberOfLettersRemoved)
  SphinxSound.playClick()
  numberOfLettersRemoved = numberOfLettersRemoved or 1
  local currentText = self.text
  assert(#currentText > 0, "Error: Tried to remove a letter when there were not any.")
  self.text = currentText:sub(1, -(numberOfLettersRemoved + 1))
  setText(self, self.text)
end

function InputDisplay:swipeSubmitted()
  self.text = ""
  -- after a word is submitted, keep it around for a bit
  local wordSubmitted = self.label:getText()
  local setTextToBlankIfTextIsTheSame = function()
    if (self.label:getText() == wordSubmitted) then
      setText(self, "")
    end
  end
  local duration = TIME_WORD_PERSISTS_AFTER_INPUT
  AHDelayed.new(duration, setTextToBlankIfTextIsTheSame):start()
end

function InputDisplay:swipeFailed()
  self.isEffectHappening = true
  local duration = 0.4
  local segmentDuration = duration/4
  local ease = nil
  local offset = self:getHeightPart(0.5)

  local shakeAnim = AHSequence.new()
  shakeAnim:addAction(AHLocationXBy.new(self, segmentDuration, ease, offset * -1))
  shakeAnim:addAction(AHLocationXBy.new(self, segmentDuration, ease, offset * 1.6))
  shakeAnim:addAction(AHLocationXBy.new(self, segmentDuration, ease, offset * -1))
  shakeAnim:addAction(AHLocationXBy.new(self, segmentDuration, ease, offset * 0.4))

  local setTextToBlankIfTextIsTheSame = function()
    self.isEffectHappening = false
    setText(self, self.text)
  end
  AHSequence.new(shakeAnim, setTextToBlankIfTextIsTheSame):start()
end

-- ----------------------------------------------------------------------------
-- Protected API
-- ----------------------------------------------------------------------------

function InputDisplay:_getRotationAnimation()
  local duration = 0.367
  local ease = AHEaseBezier.newFlat({0.42, 0, 0.58, 1})
  local animation = AHSpawn.new()

  local rotateAmount = 125
  animation:addAction(AHRotationBy.new(self.backing, duration, ease, rotateAmount))

  animation:addAction(AHLocationTo.new(self.backing, duration, ease, -self:getX(), -self:getY()))

  local labelOpacityDuration = 0.067
  animation:addAction(AHOpacityTo.new(self.label, labelOpacityDuration, nil, 0))

  local targetScaleX = self.backing:getHeightPart(0.25) / self.backing:getWidth()
  local targetScaleY = 0.25
  animation:addAction(AHScaleTo.new(self.backing, duration, ease, targetScaleX, targetScaleY))
  animation:addAction(AHScaleTo.new(self.label, duration, ease, targetScaleX, targetScaleY))

  local opacityDelay = 0.25
  local opacityDuration = 0.67
  animation:addAction(AHDelayed.new(opacityDelay, AHOpacityTo.new(self.backing, opacityDuration, nil, 0)))

  local maxScale = 1.05
  local minScale = 0.9
  local scaleAnim = AHSequence.new(
    AHScaleTo.new(self.backing, duration/2, nil, maxScale),
    AHScaleTo.new(self.backing, duration/2, nil, minScale))
  animation:addAction(scaleAnim)

  animation:addAction(AHDelayed.new(duration, function() self.wordAdded:call() end))

  local cleanup = function()
    self.backing:setLoc(self.backingX, self.backingY)
    self.backing:setRot(0)
    self.backing:setScl(1)
    self.label:setScl(1)
  end

  return AHSequence.new(animation, cleanup)
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

local function createLayout(self, characters)
  local rotatingArmGroup = Group.new(2, self:getHeightPart(0.5))
  rotatingArmGroup:setAnchorPoint(0.5, 0)
  self:insertAtLocationFractional(rotatingArmGroup, 0.5, 0)

  local cornerRadius = HEIGHT_TO_CORNER_RADIUS_MULTIPLIER * self:getHeight()
  local backingColor = DISPLAY_COLOR
  local displayBacking = FilledRoundedRectangle.new(self:getHeight(), self:getHeight(), cornerRadius, backingColor)
  rotatingArmGroup:insertAtTop(displayBacking)

  local fontSize = self:getHeight() * (1 - PERCENT_PADDING)
  local trackingOffset = fontSize * LETTER_KERNING_FONT_SIZE_MULTIPLER
  local label = Label.new("", PrerenderedFont.new(FontFaces.MEDIUM, fontSize, characters), nil, nil, nil, trackingOffset)
  label:setColor(1, 1, 1)
  rotatingArmGroup:insertAtTop(label)

  self.backing = displayBacking
  self.backingX, self.backingY = displayBacking:getLoc()
  self.backingOpacity = backingColor[4]

  self.label = label
  self.rotatingArm = rotatingArmGroup

  self.backing:setOpacity(0)
end

-- Constructor ------------------------

--- Parameter `inputCharacters` (optional) may be used to ensure that the prerendered font supports special characters e.g. 'á' or 'ñ'
function InputDisplay.new(height, characters)
  local self = Group.new(Screen.width, height)
  Mixin.mix(self, InputDisplay)

  self.wordAdded = Event.new()
  self.actionInterrupt = ActionInterrupt.new()

  createLayout(self, characters)

  self.text = ""
  self.isEffectHappening = false

  return self
end
