--- RectangleButton

class("RectangleButton")

require("sphinx.SphinxColors")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local WIDTH = Screen.width * 0.872
local HEIGHT = WIDTH * 0.196

local COLORS = {
  BLACK = {0,0,0},
  WHITE = {1,1,1},
  TEAL = {0.03, 0.69, 0.78},
  LIGHT_TEXT_COLOR = {0.87, 0.88, 0.91}
}

local DEFAULT_STYLE = {
  BUTTON_COLOR = SphinxColors.get("primaries", "aqua"),
  BUTTON_COLOR_DOWNSTATE = SphinxColors.get("mid", "aqua"),
  BUTTON_COLOR_DEPTH = SphinxColors.get("dark", "aqua"),
  BUTTON_COLOR_TEXTURE = SphinxColors.get("dark", "aqua")
}

local SECONDARY_STYLE = {
  BUTTON_COLOR = SphinxColors.get("round1", "light"),
  BUTTON_COLOR_DOWNSTATE = SphinxColors.get("round1", "mid"),
  BUTTON_COLOR_DEPTH = SphinxColors.get("round1", "darkest"),
  BUTTON_COLOR_TEXTURE = SphinxColors.get("round1", "dark")
}

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function setStatesForStandardButton(self, style)
  self.setDownState = function()
    self.facePlate:setColor(unpack(style.BUTTON_COLOR_DOWNSTATE))
  end

  self.setUpState = function()
    self.facePlate:setColor(unpack(style.BUTTON_COLOR))
  end
end

local function setStatesForSubButton(self)
  self.setDownState = function()
    self.textLabel:setOpacity(0.5)
  end

  self.setUpState = function()
    self.textLabel:setOpacity(1)
  end
end

local function setButtonFunctionality(self)
  local onTouchDown = function()
    self.setDownState()
  end
  local onTouchUp = function()
    self.setUpState()
    if self.buttonFunction then
      self.buttonFunction()
    end
  end
  local onTouchCancel = function()
    self.setUpState()
  end

  Button.makeButton(self, onTouchDown, onTouchUp, onTouchCancel)
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function RectangleButton:setFunction(buttonFunction)
  self.buttonFunction = buttonFunction
end

function RectangleButton:getFadeAction(duration, ease, opacity)
  local fadeAction = AHSpawn.new()
  if self.textLabel then
    fadeAction:addAction(AHOpacityTo.new(self.textLabel, duration, ease, opacity))
  end
  if self.facePlate then
    fadeAction:addAction(AHOpacityTo.new(self.facePlate, duration, ease, opacity))
  end
  if self.depth then
    fadeAction:addAction(AHOpacityTo.new(self.depth, duration, ease, opacity))
  end
  if self.dropShadow then
    local dropShadowOpacity = opacity * self.dropShadow.maxOpacity
    fadeAction:addAction(AHOpacityTo.new(self.dropShadow, duration, ease, dropShadowOpacity))
  end
  if self.texture then
    fadeAction:addAction(AHOpacityTo.new(self.texture, duration, ease, opacity))
  end

  return fadeAction
end

function RectangleButton:hide()
  self:getFadeAction(0,0,0):start()
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

local function createLayoutForStandardButton(self, text, style)
  local cornerRadius = self:getHeightPart(0.08)

  local dropShadowHeightExtension = self:getHeightPart(0.125)
  local dropShadow = FilledRoundedRectangle.new(
    self:getWidth(),
    self:getHeight() + dropShadowHeightExtension,
    cornerRadius,
    COLORS.BLACK)
  dropShadow.maxOpacity = 0.25
  dropShadow:setOpacity(dropShadow.maxOpacity)
  self:insertAtCenter(dropShadow)
  dropShadow:offsetY(-dropShadowHeightExtension/2)
  self.dropShadow = dropShadow

  local depthHeightExtenstion = self:getHeightPart(0.0625)
  local depth = FilledRoundedRectangle.new(
    self:getWidth(),
    self:getHeight() + depthHeightExtenstion,
    cornerRadius,
    style.BUTTON_COLOR_DEPTH)
  self:insertAtCenter(depth)
  depth:offsetY(-depthHeightExtenstion/2)
  self.depth = depth

  local facePlate = FilledRoundedRectangle.new(
    self:getWidth(),
    self:getHeight(),
    cornerRadius,
    style.BUTTON_COLOR)
  self:insertAtCenter(facePlate)
  self.facePlate = facePlate

  local texture = Sprite.newWithName("button_texture.png")
  texture:setEffectiveDims(self:getWidth(), self:getHeight())
  texture:setColor(unpack(style.BUTTON_COLOR_TEXTURE))
  self:insertAtCenter(texture)
  self.texture = texture

  local textLabel = Label.new(text, Font.new(FontFaces.BOLD, 18))
  textLabel:setColor(unpack(COLORS.WHITE))
  self:insertAtCenter(textLabel)
  self.textLabel = textLabel
end

local function createLayoutForSubButton(self, text)
  local width = WIDTH
  local baseAttributes = {
    font = Font.new(FontFaces.MEDIUM, 18),
    color = COLORS.WHITE
  }
  local attributesForLabels = {
    b = {
      font = Font.new(FontFaces.BOLD, 18),
      color = COLORS.TEAL
    },
    sb = {
      font = Font.new(FontFaces.SEMI_BOLD, 18),
      color = COLORS.LIGHT_TEXT_COLOR
    }
  }

  local textLabel = DynamicTextArea.new(
    text,
    width,
    baseAttributes,
    attributesForLabels,
    Alignment.CENTER)
  self:insertAtCenter(textLabel)
  self.textLabel = textLabel
end

local function newInternal(text, width, height, style)
  width = width or WIDTH
  height = height or HEIGHT
  local self = Group.new(width, height)
  Mixin.mix(self, RectangleButton)

  createLayoutForStandardButton(self, text, style)

  setStatesForStandardButton(self, style)
  setButtonFunctionality(self)

  return self
end

-- Constructor ------------------------

function RectangleButton.new(text, width, height)
  return newInternal(text, width, height, DEFAULT_STYLE)
end

function RectangleButton.newSecondary(text, width, height)
  return newInternal(text, width, height, SECONDARY_STYLE)
end

function RectangleButton.newSubButton(text)
  local self = Group.new(WIDTH, HEIGHT)
  Mixin.mix(self, RectangleButton)

  createLayoutForSubButton(self, text)
  self:resizeToFitChildren()

  setStatesForSubButton(self)
  setButtonFunctionality(self)

  return self
end
