--- RoundTransitionManager

class("RoundTransitionManager")

require("sphinx.ProgressMarker")


-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

RoundTransitionManager.COLOR_TRANSITION_BACKGROUND_KEY = "transitionBackground"

local Z_INDEX = {
  CANVAS = 0,
  ART_BOARD = 10,
  BACKGROUND = 15,
  NUMBER = 20,
  FOREGROUND = 30,
  ROUND_INFO = 40
}

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function getColorsForRound(self, roundNumber)
  local adjustedRoundNumber = ((roundNumber - 1) % #self.fullColorTable) + 1
  return self.fullColorTable[roundNumber]
end

local function getUpdateCanvasAnimation(self, duration)
  local oldColor = {self.canvas:getColor()}
  local newColor = self.colors.transitionBackground
  local transitionWidth = self:getWidthPart(0.25)
  local backgroundTransition = RectangleGradient.new(transitionWidth, self:getHeight(), oldColor, newColor, true)
  local newBackground = FilledRectangle.new(self:getWidth(), self:getHeight(), newColor)

  self:insertAtRight(newBackground, Z_INDEX.CANVAS)
  local offscreenOffset = newBackground:getWidthPart(0.5) + backgroundTransition:getWidth()
  newBackground:offsetX(offscreenOffset)
  self:insertAtRight(backgroundTransition, Z_INDEX.CANVAS)
  backgroundTransition:offsetX(backgroundTransition:getWidthPart(0.5))

  self.newNumber:offsetX(offscreenOffset)
  self.foreground:offsetX(offscreenOffset)
  if (self.background ~= nil) then
    self.background:offsetX(offscreenOffset)
  end

  local animation = AHSpawn.new()
  local ease = AHEaseSimpleIn.new()
  local offset = -(self:getWidth() + backgroundTransition:getWidth())
  animation:addAction(AHLocationXBy.new(self.number, duration, ease, offset))
  animation:addAction(AHLocationXBy.new(newBackground, duration, ease, offset))
  animation:addAction(AHLocationXBy.new(backgroundTransition, duration, ease, offset))
  animation:addAction(AHLocationXBy.new(self.foreground, duration, ease, -self:getWidthPart(0.75)))
  if (self.background ~= nil) then
    animation:addAction(AHLocationXBy.new(self.background, duration, ease, -self:getWidthPart(0.75)))
  end
  animation:addAction(AHLocationXBy.new(self.newNumber, duration, ease, -self:getWidthPart(0.75)))
  local cleanup = AHSpawn.new(
    AHRemove.new(self.canvas, backgroundTransition),
    function()
      self.canvas = newBackground
      self.number = self.newNumber
    end)
  return AHSequence.new(animation, cleanup)
end

local function addNumber(self, roundNumber)
  local number = Sprite.newWithName(roundNumber .. "_gradient.png")
  number:setScl(self.foreground:getCurrentScale())
  self.artBoard:insertAtCenter(number, Z_INDEX.NUMBER)
  if roundNumber == 1 then
    number:offsetX(number:getEffectiveWidthPart(-0.25))
  end

  number:offsetY(self.spacer)

  return number
end

local function prepForShow(self, roundNumber)
  if (roundNumber > 1) then
    self.progressMarkers[roundNumber-1]:setComplete()
  end
  self.progressMarkers[roundNumber]:setCurrent()
  self.roundText:setText("Round "..roundNumber)
end

local function getInitialShowAnimation(self)
  self.canvas:setColor(unpack(self.colors.transitionBackground))
  self.number = self.newNumber
  self.number:setOpacity(0)

  local fadeInElements = AHSpawn.new()

  local roundFadeInDuration = 1
  fadeInElements:addAction(AHOpacityTo.new(self.roundText, roundFadeInDuration, nil, 1))
  fadeInElements:addAction(AHOpacityTo.new(self.number, roundFadeInDuration, nil, 1))
  for _, marker in ipairs(self.progressMarkers) do
    fadeInElements:addAction(marker:getShowAnimation(roundFadeInDuration))
  end
  local propDelay = 0.2
  fadeInElements:addAction(AHDelayed.new(propDelay, self.foreground:getIntroAnimation(1)))
  if (self.background ~= nil) then
    fadeInElements:addAction(AHDelayed.new(propDelay, self.background:getIntroAnimation(1)))
  end
  local scaleDuration = self.foreground:getHoldOnScreenScaleDuration()
  fadeInElements:addAction(AHScaleTo.new(self, scaleDuration, nil, 1.05))

  self.showAnimation = fadeInElements
  return self.showAnimation
end

local function getMidGameShowAnimation(self, roundNumber)
  self.roundText:setOpacity(1)
  for _, marker in ipairs(self.progressMarkers) do
    marker:show()
  end

  local duration = 0.5
  local ease = AHEaseBezier.newFlat({0.84,0,0.16,1})
  local scaleBackIn = AHScaleTo.new(self, duration, ease, 1)

  local backgroundTransition = getUpdateCanvasAnimation(self, 0.3)
  local propSpawn = self.foreground:getIntroAnimation(roundNumber)
  if (self.background ~= nil) then
    propSpawn:addAction(self.background:getIntroAnimation(roundNumber))
  end
  local scaleDuration = self.foreground:getHoldOnScreenScaleDuration()
  local holdOnScreen = AHScaleTo.new(self, scaleDuration, nil, 1.05)

  self.showAnimation = AHSpawn.new(
    scaleBackIn,
    AHDelayed.new(0.7, backgroundTransition),
    AHDelayed.new(0.9, propSpawn),
    AHDelayed.new(0.9, holdOnScreen)
  )

  return self.showAnimation
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function RoundTransitionManager:getShowAnimation(roundNumber)
  self.roundNumber = roundNumber
  self.colors = getColorsForRound(self, roundNumber)
  prepForShow(self, roundNumber)

  self.foreground:initialize(roundNumber, self.colors)
  if (self.background ~= nil) then
    self.background:initialize(roundNumber, self.colors)
  end
  self.newNumber = addNumber(self, roundNumber)

  local anim
  if (roundNumber == 1) then
    anim = getInitialShowAnimation(self)
  else
    anim = getMidGameShowAnimation(self, roundNumber)
  end

  return anim
end

function RoundTransitionManager:getHideAnimation()
  local setup = function()
    self.showAnimation:stop()
  end

  local duration = 0.8
  local anim = AHSpawn.new()
  local scaleEase = AHEaseBezier.newFlat({0.84,0,0.16,1})
  anim:addAction(AHScaleTo.new(self, duration, scaleEase, 6))

  anim:addAction(self.foreground:getOutroAnimation(self.roundNumber))
  if (self.background ~= nil) then
    anim:addAction(self.background:getOutroAnimation(self.roundNumber))
  end

  local fadeDuration = 0.2
  local hideAnim = AHSpawn.new()
  hideAnim:addAction(AHOpacityTo.new(self.roundText, fadeDuration, nil, 0))
  hideAnim:addAction(AHOpacityTo.new(self.foreground, fadeDuration, nil, 0))
  if (self.background ~= nil) then
    hideAnim:addAction(AHOpacityTo.new(self.background, fadeDuration, nil, 0))
  end
  for _, marker in ipairs(self.progressMarkers) do
    hideAnim:addAction(marker:getHideAnimation(fadeDuration))
  end
  local cleanup = function()
    self.foreground:clear()
    if (self.background ~= nil) then
      self.background:clear()
    end
  end
  local fullAnim = AHSequence.new(setup, anim, hideAnim, cleanup)
  return fullAnim
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

local function createLayout(self, numberOfRounds)
  local canvas = FilledRectangle.new(self:getWidth(), self:getHeight(), self.colors.transitionBackground)
  self:insertAtCenter(canvas, Z_INDEX.CANVAS)

  local artBoard = Group.new(self:getWidth(), self:getHeight())
  self:insertAtCenter(artBoard, Z_INDEX.ART_BOARD)

  local roundText = Label.new("Round 0", Font.new(FontFaces.BOLD, 34))
  artBoard:insertAtCenter(roundText, Z_INDEX.ROUND_INFO)
  local verticalOffset = -self:getHeightPart(0.25)
  roundText:offsetY(verticalOffset)
  roundText:setOpacity(0)

  local progressMarkers = {}
  local radius = self:getWidthPart(0.043)
  for i=1, numberOfRounds do
    local marker = ProgressMarker.new(radius, i)
    table.insert(progressMarkers, marker)
    artBoard:insertAtCenter(marker, Z_INDEX.ROUND_INFO)
    marker:offsetX((radius * numberOfRounds) * (-(numberOfRounds + 1)/2 + i))
    marker:offsetY(verticalOffset - roundText:getHeight() - radius)
    marker:hide()
  end

  artBoard:insertAtCenter(self.foreground, Z_INDEX.FOREGROUND)
  if (self.background ~= nil) then
    artBoard:insertAtCenter(self.background, Z_INDEX.BACKGROUND)
  end

  self.spacer = self:getHeightPart(0.08)

  self.roundText = roundText
  self.progressMarkers = progressMarkers
  self.canvas = canvas
  self.artBoard = artBoard
end

-- Constructor ------------------------

-- background is optional
function RoundTransitionManager.new(foreground, fullColorTable, background)
  local self = Group.newFullScreen()
  Mixin.mix(self, RoundTransitionManager)

  self.foreground = foreground
  self.background = background
  self.fullColorTable = fullColorTable
  self.colors = fullColorTable[1]

  local standardNumberOfRounds = 3

  createLayout(self, standardNumberOfRounds)

  return self
end
