--- SphinxGame

class("SphinxGame")

require("sphinx.EndGameScreen")
require("sphinx.TutorialCarousel")
require("sphinx.RoundTransitionManager")
require("sphinx.SphinxColors")
require("sphinx.ScoreTracker")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local ROUND_COUNT = 3

local Z_INDEX = {
  ROUND_INTERSTITIAL = 0,
  SCENE = 10,
  TUTORIAL = 20,
  END = 30
}

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function createGameElements(self)
  local scoreTracker = self:_createScoreTracker()
  self.scoreTracker = scoreTracker

  local roundInterstitial = self:_createRoundInterstitial()
  self:insertAtCenter(roundInterstitial, Z_INDEX.ROUND_INTERSTITIAL)
  self.roundInterstitial = roundInterstitial
end

local function playScaleUpAnimation(self, scaledObject)
  local duration = 0.733
  local ease = AHEaseBezier.newFlat({0.76,0,0.24,1})
  local scaleTo = AHScaleTo.new(scaledObject, duration, ease, 1)
  local delayAmount = 0.1

  scaledObject:setScl(0)

  SphinxSound.playRoundNumberFly()
  self.roundInterstitial:getHideAnimation():start()
  wait(AHDelayed.new(delayAmount, scaleTo):start())
end

local function playScaleDownAnimation(self, scaledObject)
  local duration = 0.3
  local ease = nil
  wait(AHScaleTo.new(scaledObject, duration, ease, 0):start())
end

local function showRoundInterstitial(self, roundNumber)
  SphinxSound.playRoundIntro()
  wait(self.roundInterstitial:getShowAnimation(roundNumber):start())
end

local function playRound(self, roundNumber)
  showRoundInterstitial(self, roundNumber)

  local model = self:_createModel(roundNumber)
  local scene = self:_createScene(model, roundNumber, self.scoreTracker)

  if GameSession.hasNeverWon() then
    self:_enableInGameTutorial(scene)
  end

  if GameSession.hasNeverWon() and self:_hasTutorialCarousel(roundNumber) then
    local tutorialCarousel = self:_createTutorialCarousel(roundNumber)
    self:insertAtCenter(tutorialCarousel, Z_INDEX.TUTORIAL)
    playScaleUpAnimation(self, tutorialCarousel)
    self:insertAtCenter(scene, Z_INDEX.SCENE)
    tutorialCarousel:enableAndWaitUntilDismissed()
  else
    self:insertAtCenter(scene, Z_INDEX.SCENE)
    scene:setScl(0)
    playScaleUpAnimation(self, scene)
  end

  local minimumMoveCount = self:_getMinimumMoveCount(model)
  self.scoreTracker:notifyStartRound(roundNumber, minimumMoveCount)

  scene:startRound()

  --Analytics.sendGameScreenViewEvent(roundNumber)
  waitUntil(function() return scene:roundIsOver() end)

  scene:endRound()

  if roundNumber < ROUND_COUNT then
    playScaleDownAnimation(self, scene)
    scene:removeSelf()
  end
end

local function showEndGame(self)
  local displayName = self:_getDisplayName()
  local endTime = self.scoreTracker:getTimeElapsed()
  local stats = self.scoreTracker:getGameStats()

  --Analytics.sendPostPuzzleEvents(rank, progressData)

  local endGameScreen = EndGameScreen.new(displayName, endTime, stats, ROUND_COUNT)
  self:insertAtCenter(endGameScreen, Z_INDEX.END)
  endGameScreen:show()
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function SphinxGame:gameFlow()
  createGameElements(self)

  for roundNumber = 1, ROUND_COUNT do
    SphinxColors.setRoundIndex(roundNumber)
    playRound(self, roundNumber)

    --Analytics.sendPuzzleSolvedEvent(roundNumber)
  end

  self.scoreTracker:notifyGameIsCompleted()
  showEndGame(self)
end

-- ----------------------------------------------------------------------------
-- Protected API
-- ----------------------------------------------------------------------------

function SphinxGame:_createScoreTracker()
  local scoreTracker = ScoreTracker.new()
  return scoreTracker
end

function SphinxGame:_createRoundInterstitial()
  error("not implemented")
end

function SphinxGame:_createModel(roundNumber)
  error("not implemented")
end

function SphinxGame:_createScene(model, roundNumber, scoreTracker)
  error("not implemented")
end

-- Base class does nothing by default.
function SphinxGame:_enableInGameTutorial(scene)

end

function SphinxGame:_hasTutorialCarousel(roundNumber)
  return false
end

function SphinxGame:_createTutorialCarousel(roundNumber)
  return nil
end

function SphinxGame:_getMinimumMoveCount(model)
  error("not implemented")
end

function SphinxGame:_getDisplayName()
  error("not implemented")
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

-- Constructor ------------------------

function SphinxGame.new()
  local self = BaseGame.new()
  Mixin.mix(self, SphinxGame)

  return self
end
