--- SphinxSphinxUIOverlay
-- Parts of UIOverlay.lua from the Sphinx project

class("SphinxUIOverlay")

require("sphinx.SphinxColors")
require("sphinx.SphinxSound")
require("sphinx.GenericButton")
require("sphinx.CooldownGenericButton")
require("sphinx.SphinxTimer")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

SphinxUIOverlay.HOME_ROW_LEFT = "HOME_ROW_LEFT"
SphinxUIOverlay.HOME_ROW_RIGHT = "HOME_ROW_RIGHT"

local BUTTON_SIZE = ReferenceDevice.getScaledWidth(54.0)
local BUTTON_SIZE_SMALL = ReferenceDevice.getScaledWidth(40.0)
local BUTTON_MARGIN = 12.0

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function addSoundToButtonHandler(buttonHandler, playSound)
  playSound = playSound or SphinxSound.playClick

  return function()
    playSound()
    buttonHandler()
  end
end

local function getTopInset()
  if Host.safeAreaTopInset > 0 then
    return 0
  else
    return BUTTON_MARGIN
  end
end

local function getBottomInset()
  if Host.safeAreaTopInset > 0 then
    return 0
  else
    return BUTTON_MARGIN
  end
end

local function createButton(self, buttonHandler, spriteName, playSound)
  buttonHandler = addSoundToButtonHandler(buttonHandler, playSound)
  return GenericButton.newWithSpriteLabel(spriteName, BUTTON_SIZE, self.buttonColor, buttonHandler)
end

local function addHomeRowButton(self, area, button)
  local homeRow = self.homeRows[area]
  assert(homeRow ~= nil, "Unexpected area value: " .. tostring(area))

  homeRow:insert(button)
  homeRow:alignChildren()
  table.insert(self.homeRowButtons, button)
end

local function storeWidget(self, widget, widgetName)
  assert(self[widgetName] == nil, "SphinxUIOverlay already has " .. widgetName .. " set")
  self[widgetName] = widget
end

local function getButtonFlyOutAnimation(self, button, area)
  local duration = 0.2
  local ease = AHEaseSimpleIn.new()
  local buttonX = self:transformObjectLoc(button)
  local moveDistance = -buttonX - button:getWidthPart(0.5)
  if area == SphinxUIOverlay.HOME_ROW_RIGHT then
    moveDistance = self:getWidth() - buttonX + button:getWidthPart(0.5)
  end
  local move = AHLocationXBy.new(button, duration, ease, moveDistance)
  return AHSequence.new(move, function() button:setVisible(false) end)
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

-- Adding widgets ---------------------

function SphinxUIOverlay:addPauseButton()
  local pauseButton = GenericButton.newPauseButton(BUTTON_SIZE_SMALL, self.buttonColor)
  pauseButton:setAnchorPoint(1,1)
  pauseButton:setLoc(self:getWidth() - BUTTON_MARGIN, self:getHeight() - getTopInset())

  self:insert(pauseButton)
  storeWidget(self, pauseButton, "pauseButton")

  return pauseButton
end

function SphinxUIOverlay:addHintButton(area, buttonHandler)
  buttonHandler = addSoundToButtonHandler(buttonHandler, SphinxSound.playShowHint)
  local hintButton = CooldownGenericButton.new(BUTTON_SIZE, "hint_icon.png", buttonHandler, self.buttonColor, self.buttonCooldownColor)
  addHomeRowButton(self, area, hintButton)
  storeWidget(self, hintButton, "hintButton")
  return hintButton
end

function SphinxUIOverlay:addTimer()
  local timer = SphinxTimer.new(BUTTON_SIZE_SMALL, self.buttonColor)
  timer:setAnchorPoint(0, 1)
  timer:setLoc(BUTTON_MARGIN, self:getHeight() - getTopInset())

  self:insert(timer)
  storeWidget(self, timer, "timer")

  return timer
end

-- Getting widgets --------------------

function SphinxUIOverlay:getPauseButton()
  return self.pauseButton
end

function SphinxUIOverlay:getHintButton()
  return self.hintButton
end

function SphinxUIOverlay:getBonusButton()
  return self.bonusButton
end

function SphinxUIOverlay:getTimer()
  return self.timer
end

-- Managing buttons state -------------

function SphinxUIOverlay:setTouchEnabled(isEnabled)
  for _, button in ipairs(self.homeRowButtons) do
    button:setTouchEnabled(isEnabled)
  end
end

function SphinxUIOverlay:setHomeRowButtonsActive(isActive)
  for _, button in ipairs(self.homeRowButtons) do
    button:setActive(isActive)
  end
end

function SphinxUIOverlay:setHomeRowButtonsVisible(isVisible)
  for _, button in ipairs(self.homeRowButtons) do
    button:setVisible(isVisible)
  end
end

function SphinxUIOverlay:setStartOfRoundButtonPositions()
  for _, homeRow in pairs(self.homeRows) do
    homeRow:alignChildren()
  end
end

-- Getting UI dimensions --------------

function SphinxUIOverlay:getSafeTopY()
  return self:getHeight() - BUTTON_SIZE + Screen.safeBottom
end

function SphinxUIOverlay:getSafeTopYRaw()
  return self:getSafeTopY() - Host.safeAreaTopInset
end

function SphinxUIOverlay:getSafeBottomY()
  return getBottomInset() + BUTTON_SIZE
end

function SphinxUIOverlay:getSafeBottomYRaw()
  return self:getSafeBottomY() + Screen.safeBottom
end

function SphinxUIOverlay:getButtonSize()
  return BUTTON_SIZE
end

-- Animating buttons ------------------

function SphinxUIOverlay:getButtonsFlyOutAnimation()
  local animation = AHSpawn.new()
  for area, homeRow in pairs(self.homeRows) do
    for _, button in ipairs(homeRow:getChildren()) do
      animation:addAction(getButtonFlyOutAnimation(self, button, area))
    end
  end
  return animation
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

local function createHomeRows(self)
  local y = getBottomInset()

  local homeRowLeft = Row.new(BUTTON_MARGIN)
  homeRowLeft:setAnchorPointBottomLeft()
  self:insertAtLocation(homeRowLeft, BUTTON_MARGIN, y)

  local homeRowRight = Row.new(BUTTON_MARGIN)
  homeRowRight:setAnchorPointBottomRight()
  self:insertAtLocation(homeRowRight, self:getWidth() - BUTTON_MARGIN, y)

  self.homeRows = {
    [SphinxUIOverlay.HOME_ROW_LEFT] = homeRowLeft,
    [SphinxUIOverlay.HOME_ROW_RIGHT] = homeRowRight
  }
end

-- Constructor ------------------------

function SphinxUIOverlay.new(scoreTracker)
  local self = Group.newFullScreenSafe()
  Mixin.mix(self, SphinxUIOverlay)

  local roundColors = SphinxColors.CURRENT_SCENE

  self.buttonColor = roundColors.button
  self.buttonCooldownColor = roundColors.gameBackground
  self.buttonCountColor = roundColors.buttonCount
  self.pauseButton = nil
  self.hintButton = nil
  self.bonusButton = nil
  self.homeRowButtons = {}

  createHomeRows(self)

  self:addPauseButton()

  return self
end
