--- TutorialCarousel

class("TutorialCarousel")

require("sphinx.RectangleButton")
require("sphinx.SphinxCarousel")
require("sphinx.SphinxSound")
require("utils.TextUtils")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local CAROUSEL_DOT_COLOR_NEUTRAL = {1.0, 1.0, 1.0, 0.4}
local CAROUSEL_DOT_COLOR_HIGHLIGHTED = {1.0, 1.0, 1.0, 1.0}

local CLOSE_DURATION = 0.4
local CLOSE_EASE = nil
local START_BUTTON_FADE_DURATION = 0.2

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function createFadeOutAction(group)
  return AHOpacityTo.new(group, CLOSE_DURATION, CLOSE_EASE, 0)
end

local function linkCarouselOpacityToChildren(self)
  local recursive = true
  self.carousel:linkOpacityAttrToChildren(recursive)
end

local function createFadeOutElementsAction(self)
  local fadeOutElementsAction = AHSpawn.new(
    createFadeOutAction(self.closeButton),
    createFadeOutAction(self.swipeToContinueText),
    createFadeOutAction(self.instructions),
    createFadeOutAction(self.carousel),
    self.startGameButton:getFadeAction(CLOSE_DURATION, CLOSE_EASE, 0)
  )

  return fadeOutElementsAction
end

local function dismissTutorial(self)
  linkCarouselOpacityToChildren(self)

  local closeTutorial = AHSequence.new(
    createFadeOutElementsAction(self),
    AHOpacityTo.new(self.background, CLOSE_DURATION, CLOSE_EASE, 0)
  )
  wait(closeTutorial:start())
  self:removeSelf()
  self.dismissed = true
end

local function makeCloseButton(self)
  local button = Sprite.newWithName("close_icon.png")
  local downState = function()
    button:setOpacity(0.5)
  end
  local upState = function()
    button:setOpacity(1)
  end
  local onTouchDown = function()
    downState()
  end
  local onTouchUp = function()
    SphinxSound.playClick()
    upState()
    dismissTutorial(self)
  end
  local onTouchCancel = function()
    upState()
  end
  Button.addVirtualButton(button, button:getWidth()*2, button:getHeight()*2, onTouchDown, onTouchUp, onTouchCancel)
  return button
end

local function makeCard(self, imageName, text)
  local card = Group.newFullScreen()
  local maxContentHeight = self.instructions:getYBottom() - self.carousel.progressIndicator:getYTop()
  local content = Group.new(self:getWidth(), maxContentHeight)

  local image = Sprite.newWithName(imageName)

  local botImageMargin = ReferenceDevice.getScaledHeight(163),

  content:insertAtCenter(image)
  image:offsetY((maxContentHeight/2 - botImageMargin)/3)

  local baseAttributes = {
    font = Font.new(FontFaces.BOLD, 18)
  }
  local attributesForLabels = {
    blue = {color = {0.22, 0.68, 0.82}}
  }

  local lineSpacing = TextUtils.getLineSpacing(18, 26)
  local textBox = DynamicTextArea.new(
    text,
    ReferenceDevice.getScaledWidth(294),
    baseAttributes,
    attributesForLabels,
    Alignment.CENTER,
    nil,
    lineSpacing)
  content:insertAtBottom(textBox)
  textBox:offsetY(botImageMargin/2)

  local maxImageWidth = content:getWidthPart(0.9)
  local maxImageHeight = self.instructions:getYBottom() - textBox:getYTop() - (botImageMargin * 1.5)
  local maxImageWidthRatio = maxImageWidth / image:getWidth()
  local maxImageHeightRatio = maxImageHeight / image:getHeight()
  local imageScale = math.min(1.0, maxImageWidthRatio, maxImageHeightRatio)
  image:setScl(imageScale)

  card:insertAtCenter(content)
  content:setY(self.instructions:getYBottom() - maxContentHeight/2)
  card:linkOpacityAttrToChildren(true)
  return card
end

local function showStartButton(self)
  local showButton = AHSequence.new(
    self.startGameButton:getFadeAction(START_BUTTON_FADE_DURATION, nil, 1),
    function() self.startGameButton:setTouchEnabled(true) end
  )
  self.hideButtonMutex:stop()
  self.hideButtonMutex:clear()
  self.showButtonMutex:insert(showButton)
end

local function hideStartButton(self)
  local hideButton = AHSequence.new(
    function() self.startGameButton:setTouchEnabled(false) end,
    self.startGameButton:getFadeAction(START_BUTTON_FADE_DURATION, nil, 0)
  )
  self.showButtonMutex:stop()
  self.showButtonMutex:clear()
  self.hideButtonMutex:insert(hideButton)
end

local function getOnScrollEndedCallback(self)
  return function()
    local currentStep = 1 + self.carousel.progressIndicator:getProgressStep()
    if currentStep ~= self.currentStep then
      --Analytics.sendTutorialScreenViewEvent(currentStep, self.roundNumber)
      self.currentStep = currentStep
    end
    local maxStep = self.carousel.progressIndicator:getStepCount()
    if (currentStep == maxStep) then
      showStartButton(self)
    else
      hideStartButton(self)
    end
    if (currentStep > 1) then
      AHOpacityTo.new(self.swipeToContinueText, 0.2, nil, 0):start()
    end
  end
end

local function makeCardsAndSetupCarousel(self, cardsConfiguration)
  for _, cardData in ipairs(cardsConfiguration) do
    local card = makeCard(self, cardData.image, cardData.text)
    self.carousel:addCard(card)
    card:inheritOpacity(self.carousel)
  end
  self.carousel:setTouchEnabled(true)
  self.carousel:setHorizontalScrollCallback(getOnScrollEndedCallback(self))
end

local function setCardsExceptFirstVisibleRecursive(self, visible)
  local cards = self.carousel:getCards()
  local firstCardIndex = 1

  for index, card in ipairs(cards) do
    if (index ~= firstCardIndex) then
      card:setVisibleRecursive(visible)
    end
  end
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function TutorialCarousel:enableAndWaitUntilDismissed()
  setCardsExceptFirstVisibleRecursive(self, true)
  self.carousel:setTouchEnabled(true)

  if #self.carousel:getCards() == 1 then
    showStartButton(self)
  end

  --Analytics.sendTutorialScreenViewEvent(1, self.roundNumber)
  waitUntil(function()
    return self.dismissed
  end)
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

local function createLayout(self, backgroundColor)
  local background = FilledRectangle.newFullScreen(backgroundColor)
  self:insertAtCenter(background)
  self.background = background

  local carousel = SphinxCarousel.new(CAROUSEL_DOT_COLOR_HIGHLIGHTED, CAROUSEL_DOT_COLOR_NEUTRAL)
  self:insertAtCenter(carousel)
  carousel:setProgressYPosition(ReferenceDevice.getScaledHeight(120))
  carousel:setTouchEnabled(false)
  self.carousel = carousel

  local closeButton = makeCloseButton(self)
  self:insertAtTopLeft(closeButton)
  closeButton:offsetX(closeButton:getWidthPart(0.5) + ReferenceDevice.getScaledWidth(27))
  closeButton:offsetY(ReferenceDevice.getScaledHeight(-61))
  self.closeButton = closeButton

  local instructions = Label.new("Instructions", Font.new(FontFaces.BOLD, 20))
  self:insertAtTop(instructions)
  instructions:setY(closeButton:getYCenter())
  self.instructions = instructions

  local swipeToContinueText = Label.new("Swipe to continue", Font.new(FontFaces.BOLD, 14))
  swipeToContinueText.maxOpacity = 0.7
  swipeToContinueText:setOpacity(swipeToContinueText.maxOpacity)
  self:insertAtCenter(swipeToContinueText)
  self.swipeToContinueText = swipeToContinueText

  local startGameButton = RectangleButton.new("Start round")
  startGameButton:hide()
  self:insertAtCenter(startGameButton)
  startGameButton:setTouchEnabled(false)
  startGameButton:setFunction(function()
    SphinxSound.playClick()
    dismissTutorial(self)
  end)
  self.startGameButton = startGameButton

  local buttonMargin = {
    top = ReferenceDevice.getScaledHeight(23),
    bottom = ReferenceDevice.getScaledHeight(29)
  }
  local footerHeight = self.startGameButton:getHeight() + buttonMargin.top + buttonMargin.bottom
  self.carousel:setProgressYPosition(footerHeight + self.carousel.progressIndicator:getHeightPart(0.5))
  self.startGameButton:setY(buttonMargin.bottom + self.startGameButton:getHeightPart(0.5))
  self.swipeToContinueText:setY(footerHeight / 2)
end

-- Constructor ------------------------

function TutorialCarousel.new(carouselData, roundNumber, backgroundColor)
  local self = Group.newFullScreen()
  Mixin.mix(self, TutorialCarousel)
  createLayout(self, backgroundColor)

  self.currentStep = 1
  self.dismissed = false
  self.roundNumber = roundNumber
  self.hideButtonMutex = ActionMutex.new()
  self.showButtonMutex = ActionMutex.new()

  makeCardsAndSetupCarousel(self, carouselData)
  setCardsExceptFirstVisibleRecursive(self, false)

  return self
end
