--- RepeatMistakeNotification
-- @author Created by andrewcousins on November 13, 2024.

class("RepeatMistakeNotification")

require("data.generators.LearningFocusedRoundGenerator")
require("ExperimentsConfiguration")


-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local BACKGROUND_COLOR = ColorUtils.colorHex24StringToStd('#0077BA')
local TEXT_COLOR = ColorUtils.colorHex24StringToStd('#F2FCFE')

local SPRITE_NAME = "repeat_mistake_icon.png"
local LABEL_TEXT = Localizer.get("RepeatMistake.LABEL")

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function RepeatMistakeNotification:shouldShowForConcept(conceptData)
  if (not self.experimentReportedExposed) then
    ExperimentsConfiguration.REPEAT_MISTAKE_NOTIFICATION_FLAG:reportExperimentExposed()
    self.experimentReportedExposed = true
  end
  return (ExperimentsConfiguration.REPEAT_MISTAKE_NOTIFICATION_FLAG:isActive()
    and conceptData[LearningFocusedRoundGenerator.IS_RESURFACED_MISTAKE_PROPERTY])
end

function RepeatMistakeNotification:conceptPresented(conceptData, optionalDelay)
  if self:shouldShowForConcept(conceptData) then
    self:show(optionalDelay)
  else
    self:hide(optionalDelay)
  end
end

function RepeatMistakeNotification:conceptCompleted(optionalDelay)
  self:hide(optionalDelay)
end

function RepeatMistakeNotification:show(optionalDelay)
  optionalDelay = optionalDelay or 0

  local scaleDuration = 0.333
  local scaleEase = AHEaseBezier.newFlat({0.3, 0.0, 0.0, 1.0})
  local rightSideTargetX = self:getWidth() - self.rightSide:getWidthPart(0.5)
  local scaleAnimation = AHSequence.new(
    function()
      self.scalableMiddle:setScaleX(0)
      self.rightSide:setX(self.rightSide:getWidthPart(1.5))
      self.pillContainer:setOpacity(1)
    end,
    AHSpawn.new(
      AHScaleXTo.new(self.scalableMiddle, scaleDuration, scaleEase, 1),
      AHLocationXTo.new(self.rightSide, scaleDuration, scaleEase, rightSideTargetX)
    )
  )

  local iconDuration = 0.5
  local iconEase = AHEaseBezier.newFlat({0.3, 0.0, 0.0, 1.0})
  local iconRotation = -90
  local iconAnimation = AHSequence.new(
    function()
      self.icon:setRot(-iconRotation)
    end,
    AHSpawn.new(
      AHOpacityTo.new(self.icon, iconDuration, iconEase, 1),
      AHRotationBy.new(self.icon, iconDuration, iconEase, iconRotation)
    )
  )

  local labelDuration = 0.5
  local labelEase = AHEaseBezier.newFlat({0.3, 0.0, 0.0, 1.0})
  local labelDelay = 0.067
  local labelOffset = 8
  local labelAnimation = AHDelayed.new(labelDelay, AHSequence.new(
    function()
      self.label:offsetX(-labelOffset)
    end,
    AHSpawn.new(
      AHOpacityTo.new(self.label, iconDuration, iconEase, 1),
      AHLocationXBy.new(self.label, labelDuration, labelEase, labelOffset)
    )
  ))

  AHDelayed.new(optionalDelay, AHSpawn.new(
    scaleAnimation,
    iconAnimation,
    labelAnimation
  )):start()
end

function RepeatMistakeNotification:hide(optionalDelay)
  optionalDelay = optionalDelay or 0
  local duration = 0.333
  AHDelayed.new(optionalDelay, AHSpawn.new(
    AHOpacityTo.new(self.pillContainer, duration, nil, 0),
    AHOpacityTo.new(self.icon, duration, nil, 0),
    AHOpacityTo.new(self.label, duration, nil, 0)
  )):start()
end

function RepeatMistakeNotification:setHidden()
  self.pillContainer:setOpacity(0)
  self.icon:setOpacity(0)
  self.label:setOpacity(0)
end

-- Override fade in behavior when used as a widget
function RepeatMistakeNotification:createFadeInAction()
  return AHDoNothing.new()
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

local function createIcon(self, iconSize)
  local sprite = Sprite.newWithName(SPRITE_NAME)
  sprite:setEffectiveDims(nil, iconSize)
  sprite:setColor(unpack(TEXT_COLOR))
  return sprite
end

local function createPillContainer(self)
  local width, height = self:getDims()
  local cornerRadius = math.floor(height * 0.5)
  local leftSide = FilledRoundedRectangle.new(cornerRadius, height, cornerRadius, BACKGROUND_COLOR, {FilledRoundedRectangle.CORNERS.TOP_RIGHT, FilledRoundedRectangle.CORNERS.BOTTOM_RIGHT})
  local rightSide = FilledRoundedRectangle.new(cornerRadius, height, cornerRadius, BACKGROUND_COLOR, {FilledRoundedRectangle.CORNERS.TOP_LEFT, FilledRoundedRectangle.CORNERS.BOTTOM_LEFT})
  local middle = FilledRectangle.new(width - (cornerRadius * 2), height, BACKGROUND_COLOR)
  middle:setAnchorPointLeft()
  local pillContainer = Group.new(width, height)
  pillContainer:insertAtCenter(leftSide)
  pillContainer:insertAtCenter(middle)
  pillContainer:insertAtCenter(rightSide)
  leftSide:setX(cornerRadius / 2)
  middle:setX(cornerRadius)
  rightSide:setX(width - cornerRadius / 2)
  self.scalableMiddle = middle
  self.rightSide = rightSide
  self:insertAtCenter(pillContainer)
  pillContainer:linkOpacityAttrToChildren()
  return pillContainer
end

local function createText(self)
  local font = PrerenderedFont.newWithActualSize(FontFaces.DIN_OT_MEDIUM, 13)
  local trackingOffset = 0.15
  local textLabel = Label.new(LABEL_TEXT, font, nil, nil, nil, trackingOffset)
  textLabel:setColor(unpack(TEXT_COLOR))
  return textLabel
end

-- Constructor ------------------------

function RepeatMistakeNotification.new()
  local label = createText(self)
  local height = label:getHeightPart(1.75)
  local icon = createIcon(self, height * 0.65)
  local row = Row.new()
  row:insert(icon)
  row:insertSpacer(icon:getWidthPart(0.7))
  row:insert(label)
  row:alignChildren()
  local width = row:getWidthPart(1.15)

  local self = Group.new(width, height)
  Mixin.mix(self, RepeatMistakeNotification)

  self.pillContainer = createPillContainer(self)
  self:insertAtCenter(row)

  self.label = label
  self.icon = icon

  self.experimentReportedExposed = false

  return self
end
