--- RiverBendCellView

class("RiverBendCellView")


-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local Z_INDEX = {
  BASE = 0,
  HIGHLIGHT = 10,
  ACTIVE = 20,
  COMPLETE = 30,
  LETTER = 40
}

local SPRITES = {
  BASE = "tile_base.png",
  HIGHLIGHT = "tile_highlight.png",
  ACTIVE = "tile_%d_a_%s.png",
  COMPLETE = "tile_%d_c_%s.png"
}

local DURATION_SPRITE_UPDATE = 0.2
local DURATION_COMPLETE_ANSWER = 0.5
local EASE = AHEaseBezier.newFlat({0.14, 0.71, 0.34, 1})
local MULTIPLIER_FONT_SCALE = 0.32
local MAX_FONT_SIZE = 30

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function getSpriteName(self, spriteType, colorIndex)
  return string.format(spriteType, colorIndex, self.nextLinkType)
end

local function createSprite(self, spriteType, colorIndex, zIndex)
  local sprite = Sprite.newWithName(getSpriteName(self, spriteType, colorIndex))
  sprite:setEffectiveDims(self.size, nil)
  self:insertAtCenter(sprite, zIndex)

  return sprite
end

local function animateActiveSprite(self, activeSprite)
  local scaleX, scaleY = activeSprite:getScl()

  if self.dragDirection == RiverBendCellStates.DRAG_DIRECTION.UP then
    activeSprite:setAnchorPointBottom()
    activeSprite:setY(0)
    activeSprite:setScl(scaleX, 0)
    AHScaleYTo.new(activeSprite, DURATION_SPRITE_UPDATE, EASE, scaleY):start()
  elseif self.dragDirection == RiverBendCellStates.DRAG_DIRECTION.DOWN then
    activeSprite:setAnchorPointTop()
    activeSprite:setY(self:getHeight())
    activeSprite:setScl(scaleX, 0)
    AHScaleYTo.new(activeSprite, DURATION_SPRITE_UPDATE, EASE, scaleY):start()
  elseif self.dragDirection == RiverBendCellStates.DRAG_DIRECTION.LEFT then
    activeSprite:setAnchorPointRight()
    activeSprite:setX(self:getWidth())
    activeSprite:setScl(0, scaleY)
    AHScaleXTo.new(activeSprite, DURATION_SPRITE_UPDATE, EASE, scaleX):start()
  elseif self.dragDirection == RiverBendCellStates.DRAG_DIRECTION.RIGHT then
    activeSprite:setAnchorPointLeft()
    activeSprite:setX(0)
    activeSprite:setScl(0, scaleY)
    AHScaleXTo.new(activeSprite, DURATION_SPRITE_UPDATE, EASE, scaleX):start()
  elseif self.dragDirection == RiverBendCellStates.DRAG_DIRECTION.NONE then
    activeSprite:setAnchorPointCenter()
    activeSprite:setX(self:getWidthPart(0.5))
    activeSprite:setY(self:getHeightPart(0.5))
    activeSprite:setScl(0)
    AHScaleTo.new(activeSprite, DURATION_SPRITE_UPDATE, EASE, scaleX, scaleY):start()
  end
end

local function createActiveSprite(self, colorIndex)
  local activeSprite = createSprite(self, SPRITES.ACTIVE, colorIndex, Z_INDEX.ACTIVE)

  if not self.activeSprite then
    animateActiveSprite(self, activeSprite)
  elseif self.activeSprite and self.linkType ~= self.nextLinkType then
    waitForFrames(1)
    self.activeSprite:removeSelf()
  end

  self.activeSprite = activeSprite
end

local function createCompleteSprite(self, colorIndex)
  local completeSprite = createSprite(self, SPRITES.COMPLETE, colorIndex, Z_INDEX.COMPLETE)
  completeSprite:setOpacity(0)

  local animation = AHSequence.new()
  animation:addAction(AHOpacityTo.new(completeSprite, DURATION_COMPLETE_ANSWER, nil, 1))
  -- This nil check covers the case where tutorial round cells
  -- need to be completed before an activeSprite is set.
  if self.activeSprite then
    animation:addAction(function() self.activeSprite:removeSelf() end)
  end
  animation:start()
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function RiverBendCellView:updateCell(state, colorIndex)
  if self.currentState ~= state or self.linkType ~= self.nextLinkType then
    if state == RiverBendCellStates.INACTIVE then
      if self.activeSprite then
        self.activeSprite:setAnchorPointCenter()
        self.activeSprite:setX(self:getWidthPart(0.5))
        self.activeSprite:setY(self:getHeightPart(0.5))

        local activeSprite = self.activeSprite
        self.activeSprite = nil
        AHSequence.new(
          AHScaleTo.new(activeSprite, DURATION_SPRITE_UPDATE, EASE, 0),
          function() activeSprite:removeSelf() end
        ):start()
      end
    elseif state == RiverBendCellStates.ACTIVE then
      createActiveSprite(self, colorIndex)
    elseif state == RiverBendCellStates.COMPLETE then
      createCompleteSprite(self, colorIndex)
    end

    self.currentState = state
    self.linkType = self.nextLinkType
  end
end

function RiverBendCellView:setCellLinkType(linkType)
  self.nextLinkType = linkType
end

function RiverBendCellView:getCellLinkType()
  return self.linkType
end

function RiverBendCellView:setCellDragDirection(dragDirection)
  self.dragDirection = dragDirection
end

function RiverBendCellView:getFadeLabelAnimation()
  return AHOpacityTo.new(self.letterLabel, 0.4, nil, 0)
end

function RiverBendCellView:updateCellHighlight(isActive)
  local opacity = MathUtils.round(self.highlightSprite:getOpacity())
  local nextOpacity = isActive and 0.85 or 0

  if opacity ~= nextOpacity then
    AHOpacityTo.new(self.highlightSprite, DURATION_SPRITE_UPDATE, nil, nextOpacity):start()
  end
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

local function initializeCell(self)
  local baseSprite = Sprite.newWithName(SPRITES.BASE)
  baseSprite:setEffectiveDims(self.size, nil)
  self:insertAtCenter(baseSprite, Z_INDEX.BASE)

  local highlightSprite = Sprite.newWithName(SPRITES.HIGHLIGHT)
  highlightSprite:setEffectiveDims(self.size, nil)
  self:insertAtCenter(highlightSprite, Z_INDEX.HIGHLIGHT)
  highlightSprite:setOpacity(0)

  local fontSize = math.min(self.size * MULTIPLIER_FONT_SCALE, MAX_FONT_SIZE)
  local letterLabel = Label.new(self.letter, Font.new(FontFaces.BOLD, fontSize))
  self:insertAtCenter(letterLabel, Z_INDEX.LETTER)

  self.baseSprite = baseSprite
  self.highlightSprite = highlightSprite
  self.letterLabel = letterLabel
end

-- Constructor ------------------------

function RiverBendCellView.new(letter, state, size, rowIndex, columnIndex)
  local self = Group.new(size, size)
  Mixin.mix(self, RiverBendCellView)

  self.letter = letter
  self.currentState = state
  self.size = size
  self.rowIndex = rowIndex
  self.columnIndex = columnIndex
  self.nextLinkType = nil
  self.linkType = nil
  self.dragDirection = nil

  initializeCell(self)

  return self
end
