--- RiverBendGridView

class("RiverBendGridView")

require("RiverBendCellView")


-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local START_ANIMATION_DURATION = 1.25
local BOARD_PULSE_DURATION = 0.2

local COLOR_COUNT = 9

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function RiverBendGridView:updateGrid()
  for rowIndex, row in ipairs(self.cellViews) do
    for columnIndex, cellView in ipairs(row) do
      local state = self.model:getCellState(rowIndex, columnIndex)
      cellView:updateCell(state, self.listOfColors[self.colorIndex])
    end
  end
end

function RiverBendGridView:getCellIndicesAt(x, y)
  x, y = self:worldToModel(x, y)

  local widthInCells, heightInCells = self.model:getBoardDimensions()
  local rowIndex = math.max(heightInCells - math.floor(y / self.cellSize), 1)
  local columnIndex = math.min(math.floor(x / self.cellSize) + 1, widthInCells)

  return rowIndex, columnIndex
end

function RiverBendGridView:setCellLinkType(rowIndex, columnIndex, linkType)
  local cellView = self.cellViews[rowIndex][columnIndex]
  cellView:setCellLinkType(linkType)
end

function RiverBendGridView:getCellLinkType(rowIndex, columnIndex)
  local cellView = self.cellViews[rowIndex][columnIndex]
  return cellView:getCellLinkType()
end

function RiverBendGridView:setCellDragDirection(rowIndex, columnIndex, dragDirection)
  local cellView = self.cellViews[rowIndex][columnIndex]
  cellView:setCellDragDirection(dragDirection)
end

function RiverBendGridView:incrementColorIndex()
  self.colorIndex = self.colorIndex % #self.listOfColors + 1
end

function RiverBendGridView:getStartRoundAnimation()
  local duration = START_ANIMATION_DURATION
  local ease = AHEaseExponentialOut.new()
  local animation = AHSpawn.new()

  for rowIndex, row in ipairs(self.cellViews) do
    for columnIndex, cellView in ipairs(row) do
      local sign = MathUtils.booleanToSign(MathUtils.isEven(rowIndex))
      local cellAnimation = AHSequence.new()
      cellAnimation:addAction(function() cellView:setScl(1) end)
      cellAnimation:addAction(AHLocationXTo.new(cellView, duration, ease, self.cellViewsStartingXValues[rowIndex][columnIndex]))
      --This line fixes the cellViews positioning caused by a floating point arithmetic error with AHLocation
      cellAnimation:addAction(function() cellView:setX(self.cellViewsStartingXValues[rowIndex][columnIndex]) end)
      animation:addAction(cellAnimation)
    end
  end

  return animation
end

function RiverBendGridView:getBoardPulseAnimation()
  local animation = AHSequence.new()
  local duration = BOARD_PULSE_DURATION
  local ease = nil

  animation:addAction(AHScaleTo.new(self, duration, ease, 1.2))
  animation:addAction(AHScaleTo.new(self, duration, ease, 1))

  return animation
end

function RiverBendGridView:getBoardCellsFadeAnimation()
  local animation = AHSpawn.new()
  for rowIndex, row in ipairs(self.cellViews) do
    for columnIndex, cellView in ipairs(row) do
      animation:addAction(cellView:getFadeLabelAnimation())
    end
  end

  return animation
end

function RiverBendGridView:revealCellHighlight(rowIndex, columnIndex)
  local cellView = self.cellViews[rowIndex][columnIndex]
  cellView:updateCellHighlight(true)
end

function RiverBendGridView:hideCellHighlights()
  for rowIndex, row in ipairs(self.cellViews) do
    for columnIndex, cellView in ipairs(row) do
      cellView:updateCellHighlight(false)
    end
  end
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

local function initializeColorIndex(self)
  local listSize = math.min(self.model:getClueCount(), COLOR_COUNT)
  local listOfColors = TableUtils.getNumberSequence(1, listSize)
  TableUtils.shuffleArray(listOfColors)
  self.listOfColors = listOfColors
  self.colorIndex = 1
end

local function createGrid(self)
  local widthInCells, heightInCells = self.model:getBoardDimensions()
  local grid = Grid.new(widthInCells)
  local cellViews = {}

  for rowIndex = 1, heightInCells do
    local row = {}
    for columnIndex = 1, widthInCells do
      local letter = self.model:getCellLetter(rowIndex, columnIndex)
      local state = self.model:getCellState(rowIndex, columnIndex)
      local cellView = RiverBendCellView.new(letter,
                                             state,
                                             self.cellSize,
                                             rowIndex,
                                             columnIndex)
      table.insert(row, cellView)
      grid:insert(cellView)
    end
    table.insert(cellViews, row)
  end

  grid:alignChildren()
  self:insertAtCenter(grid)

  self.grid = grid
  self.cellViews = cellViews
end

local function prepareForRoundStartAnimation(self)
  local cellViewsStartingXValues = {}
  for rowIndex, row in ipairs(self.cellViews) do
    local rowXValues = {}
    for columnIndex, cellView in ipairs(row) do
      table.insert(rowXValues, cellView:getX())
      local sign = MathUtils.booleanToSign(MathUtils.isOdd(rowIndex))
      cellView:offsetX(sign * self:getWidthPart(1.25))
      cellView:setScl(0)
    end
    table.insert(cellViewsStartingXValues, rowXValues)
  end
  self.cellViewsStartingXValues = cellViewsStartingXValues
end

-- Constructor ------------------------

function RiverBendGridView.new(maxWidth, maxHeight, model)
  local widthInCells, heightInCells = model:getBoardDimensions()
  local cellSize = math.floor(math.min(maxWidth / widthInCells, maxHeight / heightInCells))
  local width = cellSize * widthInCells
  local height = cellSize * heightInCells
  local self = Group.new(width, height)
  Mixin.mix(self, RiverBendGridView)

  self.cellSize = cellSize
  self.model = model

  initializeColorIndex(self)
  createGrid(self)
  prepareForRoundStartAnimation(self)

  return self
end
