--- RiverBendModel

class("RiverBendModel")

require("RiverBendCellStates")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function areIndicesValid(self, rowIndex, columnIndex)
  return rowIndex <= self.height and rowIndex > 0 and columnIndex <= self.width and columnIndex > 0
end

local function isClueIndexValid(self, clueIndex)
  return clueIndex <= #self.clues and clueIndex >= 1
end

local function getCell(self, rowIndex, columnIndex)
  return self.board[rowIndex][columnIndex]
end

local function setCellState(self, rowIndex, columnIndex, state)
  local cell = getCell(self, rowIndex, columnIndex)
  cell.state = state
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function RiverBendModel:setCellInactive(rowIndex, columnIndex)
  setCellState(self, rowIndex, columnIndex, RiverBendCellStates.INACTIVE)
end

function RiverBendModel:setCellActive(rowIndex, columnIndex)
  setCellState(self, rowIndex, columnIndex, RiverBendCellStates.ACTIVE)
end

function RiverBendModel:setCellComplete(rowIndex, columnIndex)
  setCellState(self, rowIndex, columnIndex, RiverBendCellStates.COMPLETE)
end

function RiverBendModel:setClueComplete(clueIndex)
  assert(isClueIndexValid(self, clueIndex), "clueIndex is invalid")
  self.clues[clueIndex].complete = true
end

function RiverBendModel:isSolved()
  for _, clue in ipairs(self.clues) do
    if not clue.complete then
      return false
    end
  end

  return true
end

function RiverBendModel:getCellLetter(rowIndex, columnIndex)
  assert(areIndicesValid(self, rowIndex, columnIndex), "Indices are invalid")
  local cell = getCell(self, rowIndex, columnIndex)

  return cell.letter
end

function RiverBendModel:getCellClueIndex(rowIndex, columnIndex)
  assert(areIndicesValid(self, rowIndex, columnIndex), "Indices are invalid")
  local cell = getCell(self, rowIndex, columnIndex)

  return cell.clueIndex
end

function RiverBendModel:getCellAnswerIndex(rowIndex, columnIndex)
  assert(areIndicesValid(self, rowIndex, columnIndex), "Indices are invalid")
  local cell = getCell(self, rowIndex, columnIndex)

  return cell.answerIndex
end

function RiverBendModel:getCellState(rowIndex, columnIndex)
  assert(areIndicesValid(self, rowIndex, columnIndex), "Indices are invalid")
  local cell = getCell(self, rowIndex, columnIndex)

  return cell.state
end

function RiverBendModel:getClueText(clueIndex)
  assert(isClueIndexValid(self, clueIndex), "clueIndex is invalid")
  return self.clues[clueIndex].text
end

function RiverBendModel:getClueAnswer(clueIndex)
  assert(isClueIndexValid(self, clueIndex), "clueIndex is invalid")
  return self.clues[clueIndex].answer
end

function RiverBendModel:getClueAnswerLength(clueIndex)
  assert(isClueIndexValid(self, clueIndex), "clueIndex is invalid")
  return self.clues[clueIndex].answerLength
end

function RiverBendModel:isClueCompleted(clueIndex)
  assert(isClueIndexValid(self, clueIndex), "clueIndex is invalid")
  return self.clues[clueIndex].complete
end

function RiverBendModel:getClueCount()
  return #self.clues
end

function RiverBendModel:getCompletedClueCount()
  local completedClueCount = 0
  for _, clue in ipairs(self.clues) do
    if clue.complete then
      completedClueCount = completedClueCount + 1
    end
  end

  return completedClueCount
end

function RiverBendModel:getBoardDimensions()
  return self.width, self.height
end

function RiverBendModel:getInputCharacters()
  local characterArray = {}
  for _, row in ipairs(self.board) do
    for _, cell in ipairs(row) do
      table.insert(characterArray, cell.letter)
    end
  end

  return table.concat(characterArray)
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

local function getBoardDims(board)
  local width = #board[1]
  local height = #board

  return width, height
end

local function buildClueAnswers(self)
  for clueIndex, clue in ipairs(self.clues) do
    for answerIndex = 1, clue.answerLength do
      for rowIndex, row in ipairs(self.board) do
        for columnIndex, cell in ipairs(row) do
          local cellClueIndex = self:getCellClueIndex(rowIndex, columnIndex)
          local cellAnswerIndex = self:getCellAnswerIndex(rowIndex, columnIndex)
          if cellClueIndex == clueIndex and cellAnswerIndex == answerIndex then
            clue.answer = clue.answer .. self:getCellLetter(rowIndex, columnIndex)
            break
          end
        end
      end
    end
  end
end

local function initializeModel(self, board, clues)
  self.width, self.height = getBoardDims(board)
  self.board = board
  self.clues = {}

  for i = 1, #clues do
    local clue = {
      text = clues[i],
      answerLength = 0,
      answer = "",
      complete = false
    }
    table.insert(self.clues, clue)
  end

  for rowIndex, row in ipairs(self.board) do
    for columnIndex, cell in ipairs(row) do
      self:setCellInactive(rowIndex, columnIndex)
      self.clues[cell.clueIndex].answerLength = math.max(cell.answerIndex, self.clues[cell.clueIndex].answerLength)
    end
  end

  buildClueAnswers(self)
end

-- Constructor ------------------------

function RiverBendModel.new(board, clues)
  local self = {}
  Mixin.mix(self, RiverBendModel)

  initializeModel(self, board, clues)

  return self
end
