--- RiverBendScene

class("RiverBendScene")

require("sphinx.GenericButton")
require("sphinx.SphinxScene")
require("sphinx.SphinxUIOverlay")
require("sphinx.CelebrationOverlay")
require("sphinx.ClueDisplayCarousel")
require("sphinx.InputDisplay")
require("Event")

require("RiverBendCellStates")
require("RiverBendColors")
require("RiverBendGridView")
require("RiverBendHintHelper")
require("RiverBendInput")
require("RiverBendLinkTypeCalculator")
require("RiverBendWordStatus")
require("RiverBendDragTracker")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local Z_INDEX = {
  BACKGROUND = 0,
  INPUT = 10,
  CLUE_DISPLAY = 20,
  UI = 30,
  INPUT_DISPLAY = 40,
  GRID = 50,
  CELEBRATION_OVERLAY = 70
}

local DURATION_CLUE_DISPLAY = 0.4

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function getLinkTypeCellIndices(self, cellIndex)
  local cellIndices = self.dragTracker:getDraggedCellIndices(cellIndex)
  local prevCellIndices = self.dragTracker:getDraggedCellIndices(cellIndex - 1)
  local nextCellIndices = self.dragTracker:getDraggedCellIndices(cellIndex + 1)
  return {cellIndices, prevCellIndices, nextCellIndices}
end

local function setCellLinkType(self, linkTypeCellIndices)
  local cellIndices, prevCellIndices, nextCellIndices = unpack(linkTypeCellIndices)
  local linkType = RiverBendLinkTypeCalculator.getLinkType(cellIndices, prevCellIndices, nextCellIndices)
  local rowIndex, columnIndex = unpack(cellIndices)
  self.grid:setCellLinkType(rowIndex, columnIndex, linkType)
end

local function updateLinkTypes(self)
  local draggedWordLength = self.dragTracker:getDraggedWordLength()
  assert(draggedWordLength > 0, "cannot update linkTypes if draggedWord is empty")
  setCellLinkType(self, getLinkTypeCellIndices(self, draggedWordLength))

  if draggedWordLength > 1 then
    setCellLinkType(self, getLinkTypeCellIndices(self, draggedWordLength - 1))
  end
end

local function updateAddedCellDragDirection(self)
  local draggedWordLength = self.dragTracker:getDraggedWordLength()
  local cellIndices = self.dragTracker:getDraggedCellIndices(draggedWordLength)
  local rowIndex, columnIndex = unpack(cellIndices)
  local prevCellIndices = self.dragTracker:getDraggedCellIndices(draggedWordLength - 1)

  if prevCellIndices then
    local prevCellRowIndex, prevCellColumnIndex = unpack(prevCellIndices)
    local rowDiff = rowIndex - prevCellRowIndex
    local colDiff = columnIndex - prevCellColumnIndex

    if rowDiff == 1 then
      self.grid:setCellDragDirection(rowIndex, columnIndex, RiverBendCellStates.DRAG_DIRECTION.DOWN)
    elseif rowDiff == -1 then
      self.grid:setCellDragDirection(rowIndex, columnIndex, RiverBendCellStates.DRAG_DIRECTION.UP)
    elseif colDiff == 1 then
      self.grid:setCellDragDirection(rowIndex, columnIndex, RiverBendCellStates.DRAG_DIRECTION.RIGHT)
    elseif colDiff == -1 then
      self.grid:setCellDragDirection(rowIndex, columnIndex, RiverBendCellStates.DRAG_DIRECTION.LEFT)
    end
  else
    self.grid:setCellDragDirection(rowIndex, columnIndex, RiverBendCellStates.DRAG_DIRECTION.NONE)
  end
end

local function handleDraggedLengthChanged(self, removedLettersIndicesList, addedLetterIndices)
  for _, cellIndices in ipairs(removedLettersIndicesList) do
    self.model:setCellInactive(unpack(cellIndices))
  end

  if #removedLettersIndicesList > 0 then
    self.inputDisplay:swipeLetterRemoved(#removedLettersIndicesList)
  end

  if addedLetterIndices ~= nil then
    self.model:setCellActive(unpack(addedLetterIndices))
    self.inputDisplay:swipeLetterAdded(self.model:getCellLetter(unpack(addedLetterIndices)))
  end

  updateLinkTypes(self)
  updateAddedCellDragDirection(self)
  self.grid:updateGrid()
end

local function getDraggedWordString(self)
  local letterTable = {}
  local length = self.dragTracker:getDraggedWordLength()
  for i = 1, length do
    local draggedCellIndices = self.dragTracker:getDraggedCellIndices(i)
    local letter = self.model:getCellLetter(unpack(draggedCellIndices))
    table.insert(letterTable, letter)
  end

  local word = table.concat(letterTable)
  return word
end

local function processCorrectWord(self)
  local length = self.dragTracker:getDraggedWordLength()
  for i = 1, length do
    local draggedCellIndices = self.dragTracker:getDraggedCellIndices(i)
    self.model:setCellComplete(unpack(draggedCellIndices))
  end

  self.dragTracker:resetSelection()
end

local function processIncorrectWord(self)
  local length = self.dragTracker:getDraggedWordLength()
  for i = 1, length do
    local draggedCellIndices = self.dragTracker:getDraggedCellIndices(i)
    self.model:setCellInactive(unpack(draggedCellIndices))
  end

  self.dragTracker:resetSelection()
end

local function processDraggedWord(self)
  local isCorrect = true
  local word = getDraggedWordString(self)
  local firstCellClueIndex = self.model:getCellClueIndex(unpack(self.dragTracker:getDraggedCellIndices(1)))
  local clueAnswerLength = self.model:getClueAnswerLength(firstCellClueIndex)
  local draggedWordLength = self.dragTracker:getDraggedWordLength()
  local correctWord = self.model:getClueAnswer(firstCellClueIndex)

  if draggedWordLength ~= clueAnswerLength then
    isCorrect = false
  else
    for i = 1, draggedWordLength do
      local draggedCellIndices = self.dragTracker:getDraggedCellIndices(i)
      local draggedCellClueIndex = self.model:getCellClueIndex(unpack(draggedCellIndices))
      local draggedCellLetter = self.model:getCellLetter(unpack(draggedCellIndices))

      local isCellCorrectClue = (draggedCellClueIndex == firstCellClueIndex)
      local isCellCorrectLetter = (draggedCellLetter:lower() == correctWord[i]:lower())

      if not (isCellCorrectClue and isCellCorrectLetter) then
        isCorrect = false
        break
      end
    end
  end

  if isCorrect then
    processCorrectWord(self)
    return RiverBendWordStatus.new(RiverBendWordStatus.COMPLETE, word, firstCellClueIndex)
  else
    processIncorrectWord(self)
    return RiverBendWordStatus.new(RiverBendWordStatus.INCORRECT, word, firstCellClueIndex)
  end
end

local function getNextClueIndex(self, clueIndex)
  if self.model:isSolved() then
    return nil
  end

  clueIndex = clueIndex % self.model:getClueCount() + 1
  if self.model:isClueCompleted(clueIndex) then
    return getNextClueIndex(self, clueIndex)
  else
    return clueIndex
  end
end

local function resetHints(self)
  self.hintHelper:resetHintIndex()
  self.grid:hideCellHighlights()
  self.uiOverlay:getHintButton():setActive(true)
end

local function handleCompleteStatus(self, clueIndex)
  SphinxSound.playAcceptInteraction()
  self.scoreTracker:notifyMoveAction()
  self.inputDisplay:swipeSubmitted()

  self.model:setClueComplete(clueIndex)

  self.grid:updateGrid()
  self.grid:incrementColorIndex()
  self.clueDisplay:completeCard(clueIndex, getNextClueIndex(self, clueIndex))
  resetHints(self)
end

local function handleIncorrectStatus(self)
  SphinxSound.playRejectInteraction()
  self.scoreTracker:notifyMoveAction()
  self.inputDisplay:swipeSubmitted()
  self.inputDisplay:swipeFailed()

  self.grid:updateGrid()
end

local function handleSubmittedWord(self)
  if not self.dragTracker:isDraggedWordEmpty() then
    local status = processDraggedWord(self)
    if status:isComplete() then
      handleCompleteStatus(self, status:getClueIndex())
    elseif status:isIncorrect() then
      handleIncorrectStatus(self)
    else
      error("invalid dragged word status")
    end
  end
end

local function completeTutorialClue(self, completedCellIndices, completedClueIndex)
  for index, cellIndices in ipairs(completedCellIndices) do
    self.model:setCellComplete(unpack(cellIndices))
    local prevCellIndices = completedCellIndices[index - 1]
    local nextCellIndices = completedCellIndices[index + 1]
    setCellLinkType(self, {cellIndices, prevCellIndices, nextCellIndices})
  end

  self.model:setClueComplete(completedClueIndex)
  self.grid:updateGrid()
  self.grid:incrementColorIndex()
  self.clueDisplay:completeCard(completedClueIndex, nil)
end

local function getEndRoundAnimation(self)
  local boardPulseAnimation = self.grid:getBoardPulseAnimation()
  local fadeClueDisplayAnimation = self.clueDisplay:getCardsOpacityAnimation(DURATION_CLUE_DISPLAY, 0)
  local boardCellsFadeAnimation = self.grid:getBoardCellsFadeAnimation()
  local moveButtonsAnimation = self.uiOverlay:getButtonsFlyOutAnimation()
  local celebrationAnimation = self.celebrationOverlay:getCelebrationAnimation()

  return AHSequence.new(
    AHSpawn.new(boardPulseAnimation, fadeClueDisplayAnimation, boardCellsFadeAnimation),
    AHSpawn.new(moveButtonsAnimation, celebrationAnimation)
  )
end

local function enableInput(self, enabled)
  self.uiOverlay:setTouchEnabled(enabled)
  self.input:setTouchEnabled(enabled)
  self.clueDisplay:setTouchEnabled(enabled)
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function RiverBendScene:enableInGameTutorial()
  self.isTutorial = true
end

function RiverBendScene:startRound()
  wait(AHSpawn.new(
    self.grid:getStartRoundAnimation(),
    self.clueDisplay:getCardsOpacityAnimation(DURATION_CLUE_DISPLAY, 1)
  ):start())

  if self.isTutorial then
    local completedCellIndices, completedClueIndex = RiverBendTutorialRounds.getCompletedTutorialClueData(self.roundNumber)
    if completedCellIndices then
      completeTutorialClue(self, completedCellIndices, completedClueIndex)
    end
  end

  enableInput(self, true)
  self.uiOverlay:getTimer():start()
end

function RiverBendScene:endRound()
  enableInput(self, false)
  self.uiOverlay:getTimer():pause()
  self.scoreTracker:setTimeElapsed(self.uiOverlay:getTimer():getTime())
  waitFor(1)
  self.uiOverlay:setHomeRowButtonsActive(false)
  self.scoreTracker:notifyRoundIsCompleted()
  wait(getEndRoundAnimation(self):start())
end

function RiverBendScene:roundIsOver()
  return self.model:isSolved()
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

local function createGrid(self)
  local maxWidth = Screen.getSafeWidthPart(0.9)
  local maxHeight = Screen.getSafeHeightPart(0.65)
  local grid = RiverBendGridView.new(maxWidth, maxHeight, self.model, self.roundNumber)
  self:insertAtSafeAreaCenter(grid, Z_INDEX.GRID)
  self.grid = grid
end

local function createBackground(self, colors)
  local background = FilledRectangle.newFullScreen(colors.gameBackground)
  self:insertAtCenter(background, Z_INDEX.BACKGROUND)
  self.background = background
end

local function createDragTracker(self)
  local dragTracker = RiverBendDragTracker.new()

  dragTracker.lengthChanged:addListener(function(removedLettersIndicesList, addedLetterIndices)
    handleDraggedLengthChanged(self, removedLettersIndicesList, addedLetterIndices)
  end)

  self.dragTracker = dragTracker
end

local function hintButtonHandler(self)
  return function()
    --Analytics.sendPuzzleAssistUsedEvent(self.roundNumber, "hint")
    self.scoreTracker:notifyHelpRequested()
    self.hintHelper:doHint()
  end
end

local function createUIOverlay(self)
  local uiOverlay = SphinxUIOverlay.new(self.scoreTracker)
  uiOverlay:addHintButton(SphinxUIOverlay.HOME_ROW_RIGHT, hintButtonHandler(self))
  uiOverlay:addTimer()
  uiOverlay:getTimer():setInitialTime(self.scoreTracker:getTimeElapsed())

  self:insertAtSafeAreaCenter(uiOverlay, Z_INDEX.UI)
  self.uiOverlay = uiOverlay
end

local function createClueDisplay(self)
  local _, bottomOfGrid = self.grid:getWorldLoc()
  bottomOfGrid = bottomOfGrid + (Screen.safeBottom - Screen.safeTopInset) / 2
  local topOfUIOverlay = self.uiOverlay:getSafeBottomYRaw()
  local clueDisplayWidth = self:getWidthPart(0.86)
  local clueDisplayHeight = math.min((bottomOfGrid - topOfUIOverlay) * 0.85, self:getHeightPart(0.1))
  local displayYPos = (bottomOfGrid + topOfUIOverlay) / 2
  local displayXPos = Screen.safeCenterX

  local cluesText = {}
  for i = 1, self.model:getClueCount() do
    local clueText = self.model:getClueText(i)
    table.insert(cluesText, clueText)
  end

  local clueDisplay = ClueDisplayCarousel.new(clueDisplayWidth, clueDisplayHeight, cluesText)

  clueDisplay.stepChanged:addListener(function() resetHints(self) end)

  self:insertAtLocation(clueDisplay, displayXPos, displayYPos, Z_INDEX.CLUE_DISPLAY)
  self.clueDisplay = clueDisplay
end

local function createInputDisplay(self)
  local inputDisplayHeight = self:getHeightPart(0.035)
  local characters = self.model:getInputCharacters()
  local inputDisplay = InputDisplay.new(inputDisplayHeight, characters)
  self:insertAtCenter(inputDisplay, Z_INDEX.INPUT_DISPLAY)
  local _, bottomOfGrid = self.grid:getWorldLoc()
  bottomOfGrid = bottomOfGrid + (Screen.safeBottom - Screen.safeTopInset) / 2
  local gridTop = bottomOfGrid + self.grid:getHeight()
  local uiBottom = self.uiOverlay:getSafeTopY()
  inputDisplay:setY((uiBottom + gridTop) / 2)

  self.inputDisplay = inputDisplay
end

local function createInputOverlay(self)
  local input = RiverBendInput.new(self.model, self.grid, self.dragTracker)

  input.wordSubmitted:addListener(function()
    handleSubmittedWord(self)
  end)

  self:insertAtCenter(input, Z_INDEX.INPUT)
  self.input = input
end

local function createCelebrationOverlay(self)
  local celebrationOverlay = CelebrationOverlay.new(self.roundNumber)
  self:insertAtCenter(celebrationOverlay, Z_INDEX.CELEBRATION_OVERLAY)
  self.celebrationOverlay = celebrationOverlay
end

local function createHintHelper(self)
  local hintHelper = RiverBendHintHelper.new(self.model, self.grid, self.clueDisplay, self.uiOverlay:getHintButton())
  self.hintHelper = hintHelper
end

-- Constructor ------------------------

function RiverBendScene.new(width, height, model, roundNumber, scoreTracker)
  local self = SphinxScene.new(width, height)
  Mixin.mix(self, RiverBendScene)

  self.model = model
  self.roundNumber = roundNumber
  self.scoreTracker = scoreTracker
  self.isTutorial = false

  local colors = RiverBendColors.SCENE[roundNumber]

  createGrid(self)
  createBackground(self, colors)
  createDragTracker(self)
  createUIOverlay(self)
  createClueDisplay(self)
  createInputDisplay(self)
  createInputOverlay(self)
  createCelebrationOverlay(self)
  createHintHelper(self)

  enableInput(self, false)

  return self
end
